# 定义 客户端
import time
from typing import Type, TypeVar

from kaq_quant_common.api.rest.api_client_base import ApiClientBase
from kaq_quant_common.api.rest.instruction.models import (
    InstructionRequestBase,
    InstructionResponseBase,
)
from kaq_quant_common.api.rest.instruction.models.account import (
    ContractBalanceRequest,
    ContractBalanceResponse,
)
from kaq_quant_common.api.rest.instruction.models.order import (
    AllOpenOrdersRequest,
    AllOpenOrdersResponse,
    CancelOrderRequest,
    CancelOrderResponse,
    ChangeLeverageRequest,
    ChangeLeverageResponse,
    ModifyOrderRequest,
    ModifyOrderResponse,
    OrderRequest,
    OrderResponse,
)
from kaq_quant_common.api.rest.instruction.models.position import (
    QueryPositionRequest,
    QueryPositionResponse,
)
from kaq_quant_common.api.rest.instruction.models.transfer import (
    TransferRequest,
    TransferResponse,
)
from kaq_quant_common.utils import uuid_utils

R = TypeVar("R", bound=InstructionResponseBase)


class InstructionClient(ApiClientBase):

    # 重写一下make_request处理公用字段
    def _make_request(self, method: str, request: InstructionRequestBase, response_model: Type[R]) -> R:
        # 处理公用字段
        # 时间
        if request.event_time is None:
            request.event_time = int(time.time() * 1000)
        # TODO 任务id
        if request.task_id is None:
            request.task_id = f"t_{uuid_utils.generate_uuid()}"
        return super()._make_request(method, request, response_model)

    # 下单
    def order(self, request: OrderRequest) -> OrderResponse:
        return self._make_request("order", request, OrderResponse)

    # 修改订单
    def modify_order(self, request: ModifyOrderRequest) -> ModifyOrderResponse:
        return self._make_request("modify_order", request, ModifyOrderResponse)

    # 取消订单
    def cancel_order(self, request: CancelOrderRequest) -> CancelOrderResponse:
        return self._make_request("cancel_order", request, CancelOrderResponse)

    # 查询当前全部挂单
    def all_open_orders(self, request: AllOpenOrdersRequest) -> AllOpenOrdersResponse:
        return self._make_request("all_open_orders", request, AllOpenOrdersResponse)

    # 调整杠杆
    def change_leverage(self, request: ChangeLeverageRequest) -> ChangeLeverageResponse:
        return self._make_request("change_leverage", request, ChangeLeverageResponse)

    # 查询持仓
    def query_position(self, request: QueryPositionRequest) -> QueryPositionResponse:
        return self._make_request("query_position", request, QueryPositionResponse)

    # 划转
    def transfer(self, request: TransferRequest) -> TransferResponse:
        return self._make_request("transfer", request, TransferResponse)

    # 查询合约账户余额
    def contract_balance(self, request: ContractBalanceRequest) -> ContractBalanceResponse:
        return self._make_request("contract_balance", request, ContractBalanceResponse)



import time
import hashlib


def generate_hash_id(_series, time_salt=False):
    now = time.localtime()
    salt = str(now.tm_year) + ' ' + str(now.tm_mon) + ' ' + str(now.tm_mday) + ' ' + str(now.tm_hour)
    sss = [str(_series.iloc[i]) for i in range(len(_series))]
    if time_salt:
        str_values = " ".join(sss) + ' ' + salt
    else:
        str_values = " ".join(sss)

    return hashlib.sha3_256(str_values.encode('utf-8')).hexdigest()

def row_hash_id(_series, time_salt=False):
    now = time.localtime()
    salt = str(now.tm_year) + ' ' + str(now.tm_mon) + ' ' + str(now.tm_mday) + ' ' + str(now.tm_hour)
    sss = [str(_series.iloc[i]) for i in range(len(_series))]
    if time_salt:
        str_values = " ".join(sss) + ' ' + salt
    else:
        str_values = " ".join(sss)

    return hashlib.sha3_256(str_values.encode('utf-8')).hexdigest()

def dict_hash_id(_dict:dict, time_salt=False):
    now = time.localtime()
    salt = str(now.tm_year) + ' ' + str(now.tm_mon) + ' ' + str(now.tm_mday) + ' ' + str(now.tm_hour)
    sss = [f'{key}: {value}' for key, value in _dict.items()]
    if time_salt:
        str_values = " ".join(sss) + ' ' + salt
    else:
        str_values = " ".join(sss)

    return hashlib.sha3_256(str_values.encode('utf-8')).hexdigest()

def array_hash_id(_array, time_salt=False):
    now = time.localtime()
    salt = str(now.tm_year) + ' ' + str(now.tm_mon) + ' ' + str(now.tm_mday) + ' ' + str(now.tm_hour)
    sss = [str(i) for i in _array]
    if time_salt:
        str_values = " ".join(sss) + ' ' + salt
    else:
        str_values = " ".join(sss)

    return hashlib.sha3_256(str_values.encode('utf-8')).hexdigest()




import threading
import time
from typing import Callable

from kaq_quant_common.api.rest.instruction.models.order import (
    OrderInfo,
    OrderSide,
    OrderStatus,
    PositionStatus,
)
from kaq_quant_common.utils import logger_utils, uuid_utils


class OrderHelper:
    def __init__(self, ins_server):
        # 必须放在这里 延迟引入，否则会有循环引用问题
        from kaq_quant_common.api.rest.instruction.instruction_server_base import (
            InstructionServerBase,
        )

        self._server: InstructionServerBase = ins_server
        self._logger = logger_utils.get_logger(self)

        self._mysql_table_name_order = "kaq_futures_instruction_order"
        self._mysql_table_name_position = "kaq_futures_instruction_position"

    def process_order(self, order: OrderInfo, get_order_result: Callable):
        # 获取交易所
        exchange = self._server._exchange
        #
        if not self._do_process_order(exchange, order, get_order_result, True):
            # 马上执行，没有成功，开启线程执行
            thread = threading.Thread(
                target=self._do_process_order,
                args=(exchange, order, get_order_result, False),
            )
            thread.name = f"process_order_{order.instruction_id}_{exchange}_{order.symbol}_{order.order_id}"
            thread.daemon = True
            thread.start()

    def _do_process_order(
        self,
        exchange: str,
        order: OrderInfo,
        get_order_result: Callable,
        first=True,
    ):
        # 获取mysql
        mysql = self._server._mysql

        #
        ins_id = order.instruction_id
        order_id = order.order_id
        symbol = order.symbol
        side = order.side
        position_side = order.position_side

        side_str = "开仓"
        if side == OrderSide.SELL:
            side_str = "平仓"

        if first:
            self._logger.info(f"{ins_id}_{exchange}_{symbol} step 1. {side_str}挂单成功 {order_id}")

            # 步骤1.挂单成功 插入到订单记录
            # 获取当前时间-ms
            current_time = int(time.time() * 1000)

            if mysql is not None:
                status = OrderStatus.CREATE
                sql = f"""
                INSERT INTO {self._mysql_table_name_order} (ins_id, exchange, symbol, side, position_side, orig_price, orig_coin_quantity, order_id, status, create_time, last_update_time) 
                VALUES ( '{ins_id}', '{exchange}', '{symbol}', '{side.value}', '{order.position_side.value}', {order.target_price}, {order.quantity}, '{order_id}', '{status.value}', {current_time}, {current_time} );
                """
                execute_ret = mysql.execute_sql(sql, True)

        # 步骤2.查询订单状态 直到订单成交后
        while True:
            # 获取订单结果
            order_info = get_order_result()
            if order_info is not None:
                break
            # 只查询一次
            if first:
                return False
            # 等待
            time.sleep(1)

        # 步骤3.把最终持仓写进去
        # 平均成交价格 转float
        avg_price = float(order_info["avg_price"])
        # 最终成交数量 转float
        executed_qty = float(order_info["executed_qty"])
        # 计算出usdt数量
        executed_usdt = avg_price * executed_qty

        current_time = int(time.time() * 1000)

        if mysql is None:
            self._logger.warning(f"{ins_id}_{exchange}_{symbol} 仅操作，没有入库，请设置 mysql!!")
            return

        status = OrderStatus.FINISH
        # 更新写入最终信息
        sql = f"""
        UPDATE {self._mysql_table_name_order} 
        SET price = {avg_price}, coin_quantity = {executed_qty}, usdt_quantity = {executed_usdt}, status = '{status.value}', last_update_time = {current_time} 
        WHERE ins_id = '{ins_id}' AND exchange = '{exchange}' AND symbol = '{symbol}';
        """
        execute_ret = mysql.execute_sql(sql, True)

        self._logger.info(
            f"{ins_id}_{exchange}_{symbol} step 2. 订单成交 {order_id}, {side_str}价格 {avg_price}, {{side_str}}数量 {executed_qty}, {{side_str}}usdt {executed_usdt}"
        )
        if side == OrderSide.BUY:
            # 同时插入持仓表
            position_id = uuid_utils.generate_uuid()
            sql = f"""
            INSERT INTO {self._mysql_table_name_position} (id, exchange, symbol, position_side, coin_quantity, usdt_quantity, open_ins_id, open_price, open_time, status) 
            VALUES ( '{position_id}', '{exchange}', '{symbol}', '{position_side.value}', '{executed_qty}', '{executed_usdt}', '{ins_id}', '{avg_price}', {current_time}, '{PositionStatus.OPEN.value}' );
            """
            execute_ret = mysql.execute_sql(sql, True)

            self._logger.info(f"{ins_id}_{exchange}_{symbol} step 3. 创建持仓记录 {position_id}")
        else:
            # 需要找到对应的持仓记录
            sql = f"""
            SELECT * FROM {self._mysql_table_name_position}
            WHERE exchange = '{exchange}' AND symbol = '{symbol}' AND position_side = '{position_side.value}' AND status = '{PositionStatus.OPEN.value}' 
            ORDER BY open_time ASC;
            """
            execute_ret = mysql.execute_sql(sql)
            try:
                position_id = execute_ret.fetchone().id
                if position_id is not None:
                    # 更新持仓信息
                    sql = f"""
                    UPDATE {self._mysql_table_name_position} 
                    SET close_ins_id = '{ins_id}', close_price = {avg_price}, close_time = {current_time}, status = '{PositionStatus.CLOSE.value}'
                    WHERE id = '{position_id}';
                    """
                    execute_ret = mysql.execute_sql(sql, True)

                    self._logger.info(f"{ins_id}_{exchange}_{symbol} step 3. 更新持仓记录 {position_id}")
            except:
                pass

        return True



from enum import Enum
from typing import Any, Dict, Optional

from pydantic import BaseModel


# ws 消息类型
class WsMessageType(str, Enum):
    # 请求，对应RESPONSE
    REQUEST = "request"
    # 响应，对应REQUEST
    RESPONSE = "response"
    # 推送
    PUSH = "push"
    # 心跳
    PING = "ping"
    PONG = "pong"
    # 订阅
    SUBSCRIBE = "subscribe"
    UNSUBSCRIBE = "unsubscribe"
    #
    ACK = "ack"


class WsError(BaseModel):
    code: int
    message: str
    details: Optional[Dict[str, Any]] = None


class WsEnvelope(BaseModel):
    # 请求的类型
    type: WsMessageType
    # 请求/响应相关
    req_id: Optional[str] = None
    method: Optional[str] = None
    # 推送相关
    topic: Optional[str] = None
    # 负载
    payload: Optional[Dict[str, Any]] = None
    # 错误
    error: Optional[WsError] = None

    def model_dump_json(self) -> str:
        # 简单包装，确保枚举序列化为字符串
        return super().model_dump_json()



import threading
import time
from abc import abstractmethod

from kaq_quant_common.common.monitor_base import MonitorBase
from kaq_quant_common.common.ws_wrapper import WsWrapper
from kaq_quant_common.utils import logger_utils


# 封装http定时请求
class HttpMonitor(MonitorBase):
    def __init__(self, interval=5):
        super().__init__()
        # 执行间隔
        self._interval = interval
        self._logger = logger_utils.get_logger()

    def _do_start(self):
        # 开启一条线程，定时执行http请求
        self._ticker_thread_event = threading.Event()

        def http_request():
            # 上次请求时间
            last_request_time = 0
            while True:
                # 检查是否需要退出
                if self._ticker_thread_event.is_set():
                    self._logger.info("ticker thread exit")
                    break

                # 当前时间
                current_time = time.time()
                # 如果上次请求时间距离当前时间不足，等待
                if current_time - last_request_time < self._interval:
                    time.sleep(0.1)
                    continue

                #
                last_request_time = time.time()

                # self._logger.debug('tick start')
                try:
                    self._do_request()
                except Exception as e:
                    self._logger.error(f"http request error: {e}")
                # self._logger.debug('tick finish')
                #
                time.sleep(0.1)

        # 开启线程
        self._ticker_thread = threading.Thread(target=http_request)
        # 设置为守护线程
        self._ticker_thread.daemon = True
        self._ticker_thread.start()

    def _do_stop(self):
        if self._ticker_thread_event is not None:
            self._ticker_thread_event.set()

    # ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓

    @abstractmethod
    def _do_request(self):
        """
        子类实现
        """
        pass

    # ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓



from abc import abstractmethod

from kaq_quant_common.common.monitor_base import MonitorBase


# 通用的抽象类，包装一层ws操作
class WsWrapper(MonitorBase):
    def __init__(self):
        super().__init__()

    # stop 就是调用close，ws 更好理解
    def _do_stop(self):
        self._do_close()

    # 断开连接，主动关闭
    def close(self):
        self.stop()

    @abstractmethod
    def _do_close(self):
        pass



from abc import ABC
from functools import wraps
from typing import Callable, Type

from pydantic import BaseModel


def api_method(request_model: Type[BaseModel], response_model: Type[BaseModel]):
    """
    api 方法注解
    :param request_model: 请求模型
    :param response_model: 响应模型
    :return:
    """

    def decorator(func: Callable):
        # 将注解信息绑定到原始函数
        func._is_api_method = True
        func._request_model = request_model
        func._response_model = response_model

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        # 同步注解信息到包装函数，便于通过inspect发现
        wrapper._is_api_method = True
        wrapper._request_model = request_model
        wrapper._response_model = response_model

        return wrapper

    return decorator


# 定义 api 接口，暂时没啥用
class ApiInterface(ABC):
    pass



from dagster import DagsterRunStatus, RunsFilter, RunStatusSensorContext


# 检测dagster任务是否运行
def check_dagster_job_running(context: RunStatusSensorContext, job_name: str):
    # 查询最近的运行状态
    runs = context.instance.get_runs(filters=RunsFilter(job_name=job_name), limit=1)

    # STARTING , STARTED 都是运行中
    if runs and runs[0].status in [
        DagsterRunStatus.QUEUED,
        DagsterRunStatus.NOT_STARTED,
        DagsterRunStatus.MANAGED,
        DagsterRunStatus.STARTING,
        DagsterRunStatus.STARTED,
    ]:
        context.log.info(f"Job {job_name} is already running")
        return False
    return True



from pyhive import hive
import pandas as pd

class KaqQuantHiveClient:
    def __init__(self, host="localhost", port=10000, username="hive", password=None, auth="NONE", database="default"):
        """
        初始化 Hive 连接
        """
        self.conn = hive.Connection(
            host=host,
            port=port,
            username=username,
            password=password,
            auth=auth,
            database=database
        )
        self.cursor = self.conn.cursor()

    def query(self, sql):
        """
        执行查询并返回结果（list of tuples）
        """
        self.cursor.execute(sql)
        return self.cursor.fetchall()

    def query_df(self, sql):
        """
        执行查询并返回 pandas.DataFrame
        """
        return pd.read_sql(sql, self.conn)

    def insert_many(self, table, columns, values_list, partition=None):
        """
        批量插入数据
        :param table: Hive 表名
        :param columns: 插入的字段列表，如 ["id", "name", "department"]
        :param values_list: 数据列表，如 [(1, "Alice", "HR"), (2, "Bob", "IT")]
        :param partition: 可选分区，如 {"dt": "2025-09-21"}
        """
        cols = ",".join(columns)
        values_str = ",".join(
            ["(" + ",".join([f"'{str(v)}'" if isinstance(v, str) else str(v) for v in row]) + ")" 
             for row in values_list]
        )

        if partition:
            part_clause = " PARTITION (" + ",".join([f"{k}='{v}'" for k, v in partition.items()]) + ")"
        else:
            part_clause = ""

        sql = f"INSERT INTO {table}{part_clause} ({cols}) VALUES {values_str}"
        print("执行SQL:", sql)
        self.cursor.execute(sql)

    def close(self):
        """
        关闭连接
        """
        self.cursor.close()
        self.conn.close()

if __name__ == "__main__":
    '''
    CREATE TABLE employees (
        id INT,
        name STRING,
        position STRING,
        hire_date String
    )
    COMMENT 'Table for employee details'
    PARTITIONED BY (department STRING);

    INSERT INTO employees (id, name, position, hire_date, department) VALUES (6, 'Alice', 'Manager', '2020-05-12', 'HR');

    -- 插入第二条数据
    INSERT INTO employees (id, name, position, hire_date, department) VALUES (2, 'Bob', 'Software Engineer', '2018-11-05', 'Engineering');

    -- 插入第三条数据
    INSERT INTO employees (id, name, position, hire_date, department) VALUES (3, 'Charlie', 'Sales Representative', '2019-08-15', 'Sales');

    -- 插入第四条数据
    INSERT INTO employees (id, name, position, hire_date, department) VALUES (4, 'David', 'Data Analyst', '2021-01-22', 'Engineering');

    -- 插入第五条数据
    INSERT INTO employees (id, name, position, hire_date, department) VALUES (5, 'Eva', 'Marketing Specialist', '2020-02-17', 'Marketing');
    '''
    hive_client = KaqQuantHiveClient(
        host="127.0.0.1",
        port=10000,
        username="hive"
    )

    # 查询数据
    result = hive_client.query_df("SELECT * FROM employees")
    print(result)

    # 插入数据
    hive_client.insert_many(
        table="employees",
        columns=["id", "name", "position", "hire_date", "department"],
        values_list=[
            (101, "Alice", "Manager", "2025-09-21", "HR"),
            (102, "Bob", "Engineer", "2025-09-21", "IT")
        ],
    )

    hive_client.close()



from enum import Enum
from typing import Optional

from pydantic import BaseModel

from . import InstructionRequestBase, InstructionResponseBase


# 持仓方向
class PositionSide(str, Enum):
    # 看多
    LONG = "long"
    # 看空
    SHORT = "short"
    # 双向
    BOTH = "both"


class PositionInfo(BaseModel):
    # 交易对
    symbol: str
    # 持仓方向
    position: PositionSide
    # 持仓数量
    amount: float
    # 开仓价格
    open_price: float
    # 标记价格
    mark_price: float
    # 强平价格
    liquidation_price: float
    # 未实现盈亏
    unrealized_pnl: float
    # 初始保证金
    initial_margin: float
    # 维持保证金
    maintain_margin: float
    # 仓位初始保证金
    position_initial_margin: float
    # 订单初始保证金
    order_initial_margin: float
    # 更新时间
    update_time: int


# ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ 持仓


class QueryPositionRequest(InstructionRequestBase):
    # 交易对
    symbol: Optional[str] = None


class QueryPositionResponse(InstructionResponseBase):
    #
    positions: list[PositionInfo]



# 定义model
from typing import Optional

from pydantic import BaseModel


#
class InstructionRequestBase(BaseModel):
    # 时间
    event_time: Optional[int] = None
    # 任务id
    task_id: Optional[str] = None


class InstructionResponseBase(BaseModel):
    # 时间
    event_time: Optional[int] = None



import threading
import time
from abc import ABC, abstractmethod

from kaq_quant_common.utils import logger_utils, signal_utils


# 通用的抽象类
class MonitorBase(ABC):
    def __init__(self):
        # 初始化
        self.init()
        # 用来标记最后一次收到数据的时间，超过指定时间没有收到数据，认为连接断开
        self._keep_alive_deadline = 0

    # 初始化
    def init(self):
        # 执行初始化
        self._do_init()

    @abstractmethod
    def _do_init(self):
        pass

    # 开始，需要确保这个方法在op线程执行，执行这个方法会阻塞当前线程！
    def start(self, block=True):
        # 全局退出事件，用于传递终止信号
        exit_event = threading.Event()
        self._exit_event = exit_event

        logger = logger_utils.get_logger(self)

        #
        def handle_terminate_signal(signum, frame=None):
            """信号处理函数：捕获终止信号并触发退出事件"""
            logger.info(f"收到终止信号 {signum}")
            exit_event.set()
            # 优雅地停止
            try:
                self.stop()
            except Exception as e:
                pass

        signal_utils.register_signal_handler(handle_terminate_signal)

        self._do_start()

        # 需要阻塞
        if block:
            # 监听退出事件
            while not exit_event.is_set():
                time.sleep(1)
                if self._check_exit():
                    # 假装发送信号
                    handle_terminate_signal("exit")
                if self._keep_alive_deadline > 0 and time.time() > self._keep_alive_deadline:
                    logger.warning("MonitorBase 收据接收超时，需要停止 MonitorBase")
                    # 假装发送信号
                    handle_terminate_signal("recv time out")

            logger.warning("MonitorBase 线程退出")

    # 子类实现，判断连接是否断开，断开返回True
    def _check_exit(self):
        return False

    @abstractmethod
    def _do_start(self):
        pass

    # 停止
    def stop(self):
        self._do_stop()

    @abstractmethod
    def _do_stop(self):
        pass

    def keep_alive(self, deadline: int = 60):
        """
        保持连接alive，超过指定时间不活跃，认为需要停止
        """
        cur_time = time.time()
        self._keep_alive_deadline = cur_time + deadline



import traceback
from kaq_quant_common.resources.kaq_mysql_resources import KaqQuantMysqlRepository
from sqlalchemy import text


class KaqQuantMysqlInit:
    def __init__(self, host, port, user, passwd, database, charset, mysql_script_files):
        engine = KaqQuantMysqlRepository(host, port, user, passwd, database, charset).get_conn_engine()
        
        for file in mysql_script_files:
            try:
                with open(file, 'r', encoding='utf-8') as f:
                    content = f.read()
                # 创建数据库
                with engine.connect() as conn:
                    conn.execute(text(content))
            except Exception as e:
                print(f'【创建{file} - 表】失败, {str(e)} - {str(traceback.format_exc())}')
    

if __name__ == '__main__':
    pass
        


import inspect
import threading
import time

from flask import Flask, jsonify, request
from werkzeug.serving import make_server

from kaq_quant_common.api.common.api_interface import ApiInterface
from kaq_quant_common.api.common.auth import verify_http_request
from kaq_quant_common.utils import logger_utils, signal_utils


class ApiServerBase:
    """
    API服务器基类，用于动态发现和分派API方法。
    """

    def __init__(self, api: ApiInterface, host="0.0.0.0", port=5000):
        """
        初始化API服务器。

        :param api: 实现了ApiInterface的API实例。
        :param host: 服务器主机地址。
        :param port: 服务器端口。
        """
        self._app = Flask(__name__)
        self._api = api
        self._host = host
        self._port = port
        self._server = make_server(self._host, self._port, self._app)
        self._api_methods = self._discover_api_methods()
        self._logger = logger_utils.get_logger(self)

        @self._app.route("/api/<method_name>", methods=["POST"])
        def handle_request(method_name: str):
            """
            处理API请求。
            """
            # 简单鉴权（若未配置 token 则默认放行）
            ok, err = verify_http_request(request)
            if not ok:
                return jsonify({"error": err or "Unauthorized"}), 401

            if method_name not in self._api_methods:
                return jsonify({"error": f"Method '{method_name}' not found"}), 404

            method, request_model, response_model = self._api_methods[method_name]

            try:
                # 使用请求模型验证和解析JSON数据
                request_data = request_model(**request.json)
                # 调用API方法
                response_data = method(request_data)
                # 验证响应类型
                if not isinstance(response_data, response_model):
                    return jsonify({"error": "Invalid response type from method"}), 500
                # 允许子类包装响应
                response_data = self._wrap_response(response_data)
                # 返回JSON响应
                return jsonify(response_data.model_dump())
            except Exception as e:
                # 捕获请求处理过程中的异常
                return jsonify({"error": str(e)}), 400

    def _discover_api_methods(self):
        """
        发现并注册所有使用@api_method装饰器标记的API方法。
        """
        methods = {}
        for name, func in inspect.getmembers(self._api, predicate=inspect.ismethod):
            if hasattr(func, "_is_api_method") and func._is_api_method:
                methods[name] = (func, func._request_model, func._response_model)
        return methods

    # 子类用来包装响应，例如添加时间戳
    def _wrap_response(self, rsp: any):
        return rsp

    def run(self):
        """
        启动API服务器。
        """
        self._logger.info(f"Starting server on {self._host}:{self._port}")
        self._server.serve_forever()

    def shutdown(self):
        """
        关闭API服务器。
        """
        self._logger.info(f"Shutting down server on {self._host}:{self._port}")
        self._server.shutdown()

    def run_with_thread(self, block=True):
        """
        启动API服务器在一个新线程中。
        """
        self._server_thread = threading.Thread(target=self.run)
        self._server_thread.name = "ApiServerThread"
        self._server_thread.daemon = True
        self._server_thread.start()
        time.sleep(1)

        if block:
            self.wait_for_termination()

    def shutdown_with_thread(self):
        """
        关闭服务器并等待线程退出
        """
        try:
            self.shutdown()
        finally:
            if hasattr(self, "_server_thread") and self._server_thread.is_alive():
                self._server_thread.join(timeout=3)

    #
    def wait_for_termination(self):
        # 全局退出事件，用于传递终止信号
        exit_event = threading.Event()

        def handle_terminate_signal(signum, frame=None):
            """信号处理函数：捕获终止信号并触发退出事件"""
            self._logger.info(f"收到终止信号 {signum}")
            exit_event.set()
            # 优雅地停止服务器
            self.shutdown()

        # 监听信号
        signal_utils.register_signal_handler(handle_terminate_signal)

        # 监听退出事件
        while not exit_event.is_set():
            time.sleep(1)

        self._logger.warning("ApiServer 线程退出")



from enum import Enum

from . import InstructionRequestBase, InstructionResponseBase

# ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ 划转


# 划转类型
class TransferType(int, Enum):
    # 资金&现货
    FUNDING_SPOT = 1
    # 资金&合约
    FUNDING_FUTURES = 2
    # 合约&现货
    SPOT_FUTURES = 3


# 划转请求
class TransferRequest(InstructionRequestBase):
    # 划转类型
    type: TransferType
    # 资产/币种
    assets: str
    # 划转数量
    amount: float
    # 划转方向 1正向，2反向
    direction: int


# 划转响应
class TransferResponse(InstructionResponseBase):
    transfer_id: str



import threading
import time
from datetime import datetime, timedelta
from typing import List, Optional, Tuple, Union

import pandas as pd

from kaq_quant_common.api.ws.exchange.models import FundingRateEvent
from kaq_quant_common.api.ws.ws_server_base import WsServerBase
from kaq_quant_common.resources.kaq_mysql_resources import KaqQuantMysqlRepository
from kaq_quant_common.utils import logger_utils


class WsExchangeServer(WsServerBase):
    """
    模拟加密货币平台 WS 服务器：
    - 从 MySQL 按 exchange 读取资金费率历史数据
    - 从指定 start_time 开始，按事件时间差和加速倍数推送
    - 支持将推送事件的 event_time 重写为实时时间
    - 主题：
      * funding_rate.all
      * funding_rate.<symbol>
    """

    def __init__(
        self,
        exchange: str,
        mysql_host: str,
        mysql_port: int,
        mysql_user: str,
        mysql_passwd: str,
        mysql_db: str,
        charset: str = "utf8mb4",
        start_time: Union[int, float, datetime] = None,
        speed_multiplier: float = 1.0,
        use_realtime_event_time: bool = False,
        inject_sample_on_empty: bool = True,
        host: str = "0.0.0.0",
        port: int = 8767,
    ):
        super().__init__(self, host, port)
        self._logger = logger_utils.get_logger(self)
        # 平台
        self._exchange = exchange
        # 加速设置
        self._speed_multiplier = max(speed_multiplier if speed_multiplier and speed_multiplier > 0 else 1.0, 1e-6)
        # 是否使用实时时间
        self._use_realtime_event_time = use_realtime_event_time
        # 示例数据注入开关
        self._inject_sample_on_empty = inject_sample_on_empty

        # 单次增量加载的窗口大小
        self._window_size: timedelta = timedelta(minutes=10)
        # 触发预加载的阈值
        self._preload_horizon: timedelta = timedelta(minutes=5)
        # 丢弃数据的阈值
        self._retain_horizon: timedelta = timedelta(minutes=10)

        # 数据开始游标（时间） 支持毫秒时间戳或 datetime
        if isinstance(start_time, (int, float)):
            self._start_dt = datetime.fromtimestamp(start_time / 1000.0)
        elif isinstance(start_time, datetime):
            self._start_dt = start_time
        else:
            # 默认很早的时间，确保可以从最早记录开始
            self._start_dt = datetime.fromtimestamp(0)

        # DB
        self._repo = KaqQuantMysqlRepository(mysql_host, mysql_port, mysql_user, mysql_passwd, mysql_db, charset)

        # 推送控制
        self._push_thread: Optional[threading.Thread] = None
        self._stop_push = threading.Event()

        # 预拉取首批数据,开始的时候加载多一点
        initial_end = self._start_dt + self._preload_horizon * 2
        self._events: List[Tuple[datetime, FundingRateEvent]] = self._load_events_window(self._start_dt, initial_end)
        self._last_fetched_dt = initial_end
        self._logger.info(f"资金费率事件已加载: {len(self._events)} 条")
        # 添加测试数据
        if not self._events and self._inject_sample_on_empty:
            base_dt = self._start_dt if self._start_dt else datetime.fromtimestamp(0)
            sample: List[Tuple[datetime, FundingRateEvent]] = []
            for i in range(5):
                dt = base_dt if i == 0 else base_dt + pd.Timedelta(minutes=i)
                evt = FundingRateEvent(
                    event_time=int(dt.timestamp() * 1000),
                    exchange_symbol=f"{self._exchange}:BTCUSDT",
                    exchange=self._exchange,
                    symbol="BTCUSDT",
                    open_rate=0.0001,
                    close_rate=0.0001,
                    high_rate=0.0002,
                    low_rate=0.0000,
                )
                sample.append((dt, evt))
            self._events = sample
            self._logger.warning("资金费率历史为空，已注入示例数据用于推送测试")

        # 把开始时间调整为第一条数据的时间
        if self._events:
            self._start_dt = self._events[0][0]
        # 最后对齐整分钟
        self._start_dt = self._start_dt.replace(second=0, microsecond=0)

        # 实际启动时间，用于计算虚拟时间，需要对齐整分钟
        self._real_start_dt: datetime = datetime.now()
        self._real_start_dt = self._real_start_dt.replace(second=0, microsecond=0)

    # 原始全量拉取（保留以兼容旧逻辑）
    def _load_events(self) -> List[Tuple[datetime, FundingRateEvent]]:
        try:
            query = (
                "SELECT event_time, exchange_symbol, exchange, symbol, open_rate, close_rate, high_rate, low_rate, close_time, next_event_time, id, ctimestamp "
                "FROM kaq_future_fr_klines_1m "
                f"WHERE exchange = '{self._exchange}' AND event_time >= '{self._start_dt.strftime('%Y-%m-%d %H:%M:%S')}' "
                "ORDER BY event_time ASC"
            )
            df: pd.DataFrame = self._repo.fetch_data(query)
            if df is None or df.empty:
                self._logger.warning("资金费率表无数据或查询为空")
                return []

            events: List[Tuple[datetime, FundingRateEvent]] = []
            for _, row in df.iterrows():
                evt_dt: datetime = row.get("event_time")
                if not isinstance(evt_dt, datetime):
                    try:
                        evt_dt = pd.to_datetime(evt_dt).to_pydatetime()
                    except Exception:
                        continue
                # 转为毫秒时间戳
                close_time_val = row.get("close_time")
                next_event_time_val = row.get("next_event_time")
                try:
                    close_time_ms = int(pd.to_datetime(close_time_val).timestamp() * 1000) if close_time_val is not None else None
                except Exception:
                    close_time_ms = None
                try:
                    next_event_time_ms = int(pd.to_datetime(next_event_time_val).timestamp() * 1000) if next_event_time_val is not None else None
                except Exception:
                    next_event_time_ms = None
                evt_ms = int(evt_dt.timestamp() * 1000)
                event = FundingRateEvent(
                    event_time=evt_ms,
                    exchange_symbol=row.get("exchange_symbol"),
                    exchange=row.get("exchange"),
                    symbol=row.get("symbol"),
                    open_rate=row.get("open_rate"),
                    close_rate=row.get("close_rate"),
                    high_rate=row.get("high_rate"),
                    low_rate=row.get("low_rate"),
                    close_time=close_time_ms,
                    next_event_time=next_event_time_ms,
                    #
                    id=row.get("id"),
                    ctimestamp=row.get("ctimestamp"),
                )
                events.append((evt_dt, event))
            return events
        except Exception as e:
            self._logger.error(f"加载资金费率事件失败: {e}")
            return []

    # 增量批量拉取（保留以兼容旧逻辑）
    def _load_events_batch(self, start_from: datetime, limit: int) -> List[Tuple[datetime, FundingRateEvent]]:
        try:
            query = (
                "SELECT event_time, exchange_symbol, exchange, symbol, open_rate, close_rate, high_rate, low_rate, close_time, next_event_time, id, ctimestamp "
                "FROM kaq_future_fr_klines_1m "
                f"WHERE exchange = '{self._exchange}' AND event_time > '{start_from.strftime('%Y-%m-%d %H:%M:%S')}' "
                "ORDER BY event_time ASC "
                f"LIMIT {int(limit)}"
            )
            df: pd.DataFrame = self._repo.fetch_data(query)
            if df is None or df.empty:
                return []
            events: List[Tuple[datetime, FundingRateEvent]] = []
            for _, row in df.iterrows():
                evt_dt: datetime = row.get("event_time")
                if not isinstance(evt_dt, datetime):
                    try:
                        evt_dt = pd.to_datetime(evt_dt).to_pydatetime()
                    except Exception:
                        continue
                # 转为毫秒时间戳
                close_time_val = row.get("close_time")
                next_event_time_val = row.get("next_event_time")
                try:
                    close_time_ms = int(pd.to_datetime(close_time_val).timestamp() * 1000) if close_time_val is not None else None
                except Exception:
                    close_time_ms = None
                try:
                    next_event_time_ms = int(pd.to_datetime(next_event_time_val).timestamp() * 1000) if next_event_time_val is not None else None
                except Exception:
                    next_event_time_ms = None
                evt_ms = int(evt_dt.timestamp() * 1000)
                event = FundingRateEvent(
                    event_time=evt_ms,
                    exchange_symbol=row.get("exchange_symbol"),
                    exchange=row.get("exchange"),
                    symbol=row.get("symbol"),
                    open_rate=row.get("open_rate"),
                    close_rate=row.get("close_rate"),
                    high_rate=row.get("high_rate"),
                    low_rate=row.get("low_rate"),
                    close_time=close_time_ms,
                    next_event_time=next_event_time_ms,
                    #
                    id=row.get("id"),
                    ctimestamp=row.get("ctimestamp"),
                )
                events.append((evt_dt, event))
            return events
        except Exception as e:
            self._logger.error(f"增量加载资金费率事件失败: {e}")
            return []

    # 按照时间窗口拉取
    def _load_events_window(self, start_from: datetime, end_until: datetime) -> List[Tuple[datetime, FundingRateEvent]]:
        try:
            query = (
                "SELECT event_time, exchange_symbol, exchange, symbol, open_rate, close_rate, high_rate, low_rate, close_time, next_event_time, id, ctimestamp "
                "FROM kaq_future_fr_klines_1m "
                f"WHERE exchange = '{self._exchange}' AND event_time > '{start_from.strftime('%Y-%m-%d %H:%M:%S')}' AND event_time <= '{end_until.strftime('%Y-%m-%d %H:%M:%S')}' "
                "ORDER BY event_time ASC"
            )
            df: pd.DataFrame = self._repo.fetch_data(query)
            if df is None or df.empty:
                return []
            events: List[Tuple[datetime, FundingRateEvent]] = []
            for _, row in df.iterrows():
                evt_dt: datetime = row.get("event_time")
                if not isinstance(evt_dt, datetime):
                    try:
                        evt_dt = pd.to_datetime(evt_dt).to_pydatetime()
                    except Exception:
                        continue
                close_time_val = row.get("close_time")
                next_event_time_val = row.get("next_event_time")
                try:
                    close_time_ms = int(pd.to_datetime(close_time_val).timestamp() * 1000) if close_time_val is not None else None
                except Exception:
                    close_time_ms = None
                try:
                    next_event_time_ms = int(pd.to_datetime(next_event_time_val).timestamp() * 1000) if next_event_time_val is not None else None
                except Exception:
                    next_event_time_ms = None
                evt_ms = int(evt_dt.timestamp() * 1000)
                event = FundingRateEvent(
                    event_time=evt_ms,
                    exchange_symbol=row.get("exchange_symbol"),
                    exchange=row.get("exchange"),
                    symbol=row.get("symbol"),
                    open_rate=row.get("open_rate"),
                    close_rate=row.get("close_rate"),
                    high_rate=row.get("high_rate"),
                    low_rate=row.get("low_rate"),
                    close_time=close_time_ms,
                    next_event_time=next_event_time_ms,
                    #
                    id=row.get("id"),
                    ctimestamp=row.get("ctimestamp"),
                )
                events.append((evt_dt, event))
            return events
        except Exception as e:
            self._logger.error(f"时间窗加载资金费率事件失败: {e}")
            return []

    # 从数据库拉取最新的记录
    def _get_latest_event_dt(self) -> Optional[datetime]:
        try:
            query = "SELECT MAX(event_time) AS max_event_time " "FROM kaq_future_fr_klines_1m " f"WHERE exchange = '{self._exchange}'"
            df: pd.DataFrame = self._repo.fetch_data(query)
            if df is None or df.empty:
                return None
            val = df.iloc[0].get("max_event_time")

            if isinstance(val, datetime):
                return val
            try:
                return pd.to_datetime(val).to_pydatetime()
            except Exception:
                return None
        except Exception as e:
            self._logger.error(f"查询最新资金费率事件时间失败: {e}")
            return None

    def _push_loop(self):
        # 初始缓冲检查
        if not self._events:
            # 获取一次最新记录
            latest_dt = self._get_latest_event_dt()
            initial_end = self._last_fetched_dt + self._window_size
            # 防止超过
            if latest_dt is not None and initial_end > latest_dt:
                initial_end = latest_dt
            self._events = self._load_events_window(self._last_fetched_dt, initial_end)
            # 主要是为了记录这个最新拉取的时间点，下一次增量加载从这里开始
            self._last_fetched_dt = initial_end
        if not self._events:
            self._logger.warning("无可推送的资金费率历史事件")
            return

        prev_dt = None
        prev_counter = 0
        discard_dt = None
        idx = 0
        while not self._stop_push.is_set():
            # 虚拟时钟，考虑加速比
            now_real = datetime.now()
            # 计算出虚拟时间点
            now_virtual = self._start_dt + (now_real - self._real_start_dt) * self._speed_multiplier

            # 确保缓冲区覆盖到虚拟时间的预加载阈值
            # 获取缓冲区最新的记录时间
            buffer_max_dt = self._events[-1][0] if self._events else self._last_fetched_dt
            # 根据虚拟时间点计算出需要覆盖的时间点
            target_cover_dt = now_virtual + self._preload_horizon
            # 最新的记录时间
            latest_dt: datetime = None
            while buffer_max_dt < target_cover_dt and not self._stop_push.is_set():
                # 每次进入循环，先拉取一次最新数据时间
                if latest_dt is None:
                    latest_dt = self._get_latest_event_dt()
                # 若数据库已无更晚数据，停止增量加载
                if latest_dt is not None and self._last_fetched_dt >= latest_dt:
                    buffer_max_dt = latest_dt
                    self._logger.debug("已到达数据库最新事件时间，停止增量加载")
                    break
                # 计算出增量加载的窗口时间点
                window_end = self._last_fetched_dt + self._window_size
                # 限制窗口终点不超过数据库最新时间
                if latest_dt is not None and window_end > latest_dt:
                    window_end = latest_dt
                # 增量加载
                new_events = self._load_events_window(self._last_fetched_dt, window_end)
                self._last_fetched_dt = window_end
                if new_events:
                    self._logger.debug(f"增量获取{len(new_events)}条数据,截止时间{window_end}")
                    self._events.extend(new_events)
                    buffer_max_dt = self._events[-1][0]
                else:
                    # 即便当前窗无数据，推进游标，避免卡住
                    buffer_max_dt = max(buffer_max_dt, window_end)
                    self._logger.debug(f"增量获取{0}条数据,截止时间{window_end}")
            latest_dt = None

            # 如果缓冲区已消费完，退出（或等待下一轮覆盖）
            if idx >= len(self._events):
                time.sleep(0.05)
                continue

            # 取出一条数据
            evt_dt, evt = self._events[idx]

            # 计算等待时间，按加速比缩放
            try:
                # 数据时间 - 当前虚拟时间
                delta = (evt_dt - now_virtual).total_seconds()
                # 等待秒数
                wait_s = max(delta, 0.0)
                if wait_s > 0:
                    time.sleep(0.05)
                    continue

                if prev_dt != evt_dt:
                    # 输出一下时间
                    if prev_dt is not None:
                        self._logger.debug(f"完成处理{prev_dt}的资金费率{prev_counter}条")
                    self._logger.debug(f"开始处理{evt_dt}的资金费率")
                    prev_counter = 0
                    prev_dt = evt_dt
                    #
                    if discard_dt is None:
                        discard_dt = prev_dt
            except Exception:
                pass

            #
            idx += 1
            prev_counter += 1

            # 构造推送数据
            payload = evt.model_dump()
            if self._use_realtime_event_time:
                # 调整为实时时间
                payload["event_time"] = int(time.time() * 1000)
            # 广播到主题：all 和 symbol
            try:
                self.broadcast("funding_rate.all", payload)
                sym = payload.get("symbol") or ""
                if sym:
                    self.broadcast(f"funding_rate.{sym}", payload)
            except Exception as e:
                self._logger.warning(f"广播资金费率事件失败: {e}")

            # 时间型丢弃：移除过旧数据以控制内存
            if prev_dt - discard_dt > self._retain_horizon:
                discard_dt = prev_dt
                cut_idx = -1
                for i in range(len(self._events)):
                    if self._events[i][0] < discard_dt:
                        cut_idx = i
                    elif self._events[i][0] >= discard_dt:
                        break
                # 裁剪
                if cut_idx > 0:
                    cut_count = cut_idx + 1
                    self._logger.debug(f"丢弃{cut_count}条过期数据")
                    self._events = self._events[cut_count:]
                    # 修正idx
                    idx = max(0, idx - cut_count)

    def run_with_thread(self, block: bool = True):
        # 启动推送线程
        self._stop_push.clear()
        self._push_thread = threading.Thread(target=self._push_loop, name="WsExchangePush", daemon=True)
        self._push_thread.start()
        # 启动WS服务器线程
        super().run_with_thread(block)

    def shutdown_with_thread(self):
        # 先停止推送线程
        try:
            self._stop_push.set()
            if self._push_thread and self._push_thread.is_alive():
                self._push_thread.join(timeout=3)
        except Exception:
            pass
        # 再关闭WS服务器
        super().shutdown_with_thread()

    async def _ws_handler(self, ws):
        # 覆盖基类鉴权，模拟平台测试环境不启用鉴权
        try:
            await self._handle_connection(ws)
        except Exception:
            pass



import json

from kaq_quant_common.common.redis_table_monitor import RedisTableMonitor
from kaq_quant_common.utils import yml_utils


# 订单簿交易对监听器
class LimitOrderSymbolMonitor(RedisTableMonitor):
    def __init__(self, exchange: str, handler: callable = None, default_symbols=["BTCUSDT", "ETHUSDT"]):
        #
        self._exchange = exchange
        # 回调
        self._handler = handler
        #
        self._support_symbols = default_symbols
        # 记录一下之前的字符串
        self._str_value = None

        # 间隔1秒
        super().__init__(table_name="kaq_all_futures_limit_order_symbols_config", interval=1)

        # 输出一下支持的交易对
        # self._logger.info(f"init {self._exchange} limit order support symbols: {self._support_symbols}")

    def _on_get_redis_config(self):
        # 获取redis配置
        host, port, passwd = yml_utils.get(f"kaq_{self._exchange}_quant", "redis", ["host", "port", "passwd"])
        return {"host": host, "port": int(port), "passwd": passwd}

    def _do_query(self) -> str:
        # 获取字符串
        str = super()._do_query()
        try:
            # 解析为json
            json_obj = json.loads(str)
            # 只要平台的值
            symbols = json_obj.get(self._exchange, self._support_symbols)

            # !! 注意，这里需要过滤一下，只保留支持的交易对
            tmp_symbols = [symbol for symbol in symbols if symbol in self._support_symbols]
            # df_symbols = pd.Series(symbols)
            # df_symbols = df_symbols[df_symbols.isin(self._support_symbols)]
            # symbols = df_symbols.tolist()
            if tmp_symbols != symbols and self._str_value != str:
                # 输出一下日志
                self._logger.warning(f"{self._exchange} limit order symbols contain unsupported symbols, before: {symbols} filtered: {tmp_symbols}")

            symbols = tmp_symbols
        except Exception as e:
            symbols = self._support_symbols

        self._str_value = str

        return symbols

    def _do_compare(self, value1: list[str], value2: list[str]) -> bool:
        # 判断数组是否一样，数组也是这样判断也可以
        return value1 == value2

    def _on_value_change(self, value: list[str]):
        # 输出一下变化的交易对
        self._logger.info(f"{self._exchange} limit order symbols changed: {value}")
        if self._handler is not None:
            self._handler(value)

    # get, set
    def set_handler(self, handler: callable):
        self._handler = handler

    def get_symbols(self) -> list[str]:
        return self._value



import asyncio
import threading
from typing import Callable, Dict, Optional, Type, TypeVar

import websockets
from pydantic import BaseModel

from kaq_quant_common.api.ws.models import WsEnvelope, WsMessageType
from kaq_quant_common.utils import logger_utils, uuid_utils
from kaq_quant_common.api.common.auth import get_auth_token

R = TypeVar("R", bound=BaseModel)


class WsClientBase:
    """
    WebSocket 客户端基类：
    - 管理长连接、自动重连和心跳（基于自定义 ping/pong）
    - 请求/响应关联（req_id -> Future）
    - 订阅主题并接收服务器推送
    提供同步方法封装，便于调用。
    """

    def __init__(self, url: str, auto_reconnect: bool = True, token: Optional[str] = None):
        self._url = url
        self._auto_reconnect = auto_reconnect
        self._token = token if token is not None else get_auth_token()
        self._logger = logger_utils.get_logger(self)

        # 事件循环线程
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._loop_thread: Optional[threading.Thread] = None

        # 连接与控制
        self._ws: Optional[websockets.WebSocketClientProtocol] = None
        self._connected_event = threading.Event()
        self._stop_flag = threading.Event()

        # 请求映射
        self._pending_requests: Dict[str, asyncio.Future] = {}

        # 订阅映射：topic -> handler
        self._subscriptions: Dict[str, Callable[[dict], None]] = {}

        # 重连参数
        self._reconnect_initial = 1.0
        self._reconnect_max = 30.0

    # ============================== 外部API ==============================
    def connect(self):
        """启动事件循环与连接任务"""
        if self._loop_thread and self._loop_thread.is_alive():
            return

        # 开一条线程跑事件循环
        def _run_loop():
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)
            try:
                self._loop.run_until_complete(self._connect_forever())
            finally:
                try:
                    self._loop.close()
                except Exception:
                    pass
                self._logger.warning("WsClient 事件循环结束")

        # 循环接收线程，设置为守护进程，主线程关闭就自动关闭
        self._loop_thread = threading.Thread(target=_run_loop, name="WsClientLoop", daemon=True)
        self._loop_thread.start()

    # 断开连接
    def disconnect(self):
        self._stop_flag.set()
        if self._loop and self._ws:

            async def _close():
                # 关闭ws
                try:
                    await self._ws.close()
                except Exception:
                    pass

            try:
                fut = asyncio.run_coroutine_threadsafe(_close(), self._loop)
                # 尝试快速关闭，若事件循环已结束则忽略超时
                fut.result(timeout=1)
            except Exception:
                pass
        # 等待线程结束
        if self._loop_thread and self._loop_thread.is_alive():
            self._loop_thread.join(timeout=5)

    # 订阅，向服务器注册订阅
    def subscribe(self, topic: str, handler: Callable[[dict], None]):
        """注册订阅，并在连接后自动发送SUBSCRIBE"""
        self._subscriptions[topic] = handler
        if self._loop and self._ws:

            async def _send_sub():
                # 构造订阅消息
                env = WsEnvelope(type=WsMessageType.SUBSCRIBE, topic=topic)
                await self._ws.send(env.model_dump_json())

            asyncio.run_coroutine_threadsafe(_send_sub(), self._loop)

    # 取消订阅
    def unsubscribe(self, topic: str):
        self._subscriptions.pop(topic, None)
        if self._loop and self._ws:

            async def _send_unsub():
                # 构造取消订阅消息
                env = WsEnvelope(type=WsMessageType.UNSUBSCRIBE, topic=topic)
                # 发送取消订阅消息
                await self._ws.send(env.model_dump_json())

            asyncio.run_coroutine_threadsafe(_send_unsub(), self._loop)

    #
    def send_request(self, method: str, request_data: BaseModel, response_model: Type[R], timeout: float = 10.0) -> R:
        """同步请求封装，等待响应返回"""
        if not self._connected_event.wait(timeout=5.0):
            raise RuntimeError("WS not connected")

        async def _send_and_wait() -> R:
            # 生成请求id，用来处理响应
            req_id = f"r_{uuid_utils.generate_uuid()}"
            # 在事件循环中创建Future
            fut = self._loop.create_future()
            # 记录正在处理的请求
            self._pending_requests[req_id] = fut
            # 构造请求消息
            env = WsEnvelope(type=WsMessageType.REQUEST, req_id=req_id, method=method, payload=request_data.model_dump())
            # 发送
            await self._ws.send(env.model_dump_json())
            try:
                # 等待执行完毕
                payload: dict = await asyncio.wait_for(fut, timeout=timeout)
                # 构造返回结构
                return response_model(**payload)
            finally:
                # 移除已处理请求
                self._pending_requests.pop(req_id, None)

        # 提交到事件循环执行
        cfut = asyncio.run_coroutine_threadsafe(_send_and_wait(), self._loop)
        return cfut.result()

    # ============================== 内部协程 ==============================
    async def _connect_forever(self):
        backoff = self._reconnect_initial
        while not self._stop_flag.is_set():
            try:
                self._logger.info(f"WS connecting to {self._url}")
                headers = {}
                if self._token:
                    headers["Authorization"] = f"Bearer {self._token}"
                self._ws = await websockets.connect(self._url, ping_interval=None, additional_headers=headers or None)
                self._connected_event.set()
                self._logger.info("WS connected")
                # 连接恢复后自动订阅
                await self._resubscribe_all()
                # 启动接收循环
                await self._recv_loop()
                # 正常退出接收循环（例如主动断开）
                if self._stop_flag.is_set():
                    break
            except Exception as e:
                self._connected_event.clear()
                self._logger.warning(f"WS 连接失败或中断: {e}")
                if not self._auto_reconnect or self._stop_flag.is_set():
                    break
                await asyncio.sleep(backoff)
                backoff = min(backoff * 2, self._reconnect_max)

        # 退出时确保关闭连接
        try:
            if self._ws:
                await self._ws.close()
        except Exception:
            pass

    # 重新自动订阅
    async def _resubscribe_all(self):
        for topic in list(self._subscriptions.keys()):
            try:
                env = WsEnvelope(type=WsMessageType.SUBSCRIBE, topic=topic)
                await self._ws.send(env.model_dump_json())
            except Exception as e:
                self._logger.warning(f"自动订阅 {topic} 失败: {e}")

    # 循环接收
    async def _recv_loop(self):
        assert self._ws is not None
        try:
            async for message in self._ws:
                # 收到消息会进入到这里
                env = None
                try:
                    # 校验消息格式是否正确
                    env = WsEnvelope.model_validate_json(message)
                except Exception as e:
                    self._logger.error(f"解析消息失败: {e}")
                    continue

                if env.type == WsMessageType.RESPONSE:
                    # 对应REQUEST的响应
                    req_id = env.req_id or ""
                    # 找到对应的请求Future
                    fut = self._pending_requests.get(req_id)
                    # 如果Future存在且未完成
                    if fut and not fut.done():
                        if env.error:
                            # 异常响应，设置Future异常
                            fut.set_exception(RuntimeError(env.error.get("message") if isinstance(env.error, dict) else str(env.error)))
                        else:
                            # 成功响应，设置Future结果
                            fut.set_result(env.payload or {})
                elif env.type == WsMessageType.PUSH:
                    # 推送过来的订阅的消息
                    topic = env.topic or ""
                    # 找到对应订阅处理的函数
                    handler = self._subscriptions.get(topic)
                    if handler and env.payload is not None:
                        # 在后台执行handler，避免阻塞事件循环
                        asyncio.create_task(asyncio.to_thread(handler, env.payload))
                elif env.type == WsMessageType.PING:
                    # 心跳
                    pong = WsEnvelope(type=WsMessageType.PONG)
                    await self._ws.send(pong.model_dump_json())
        except Exception as e:
            self._logger.warning(f"WS 接收循环异常: {e}")
        finally:
            self._connected_event.clear()



import threading
import time
from abc import abstractmethod

from kaq_quant_common.common.monitor_base import MonitorBase
from kaq_quant_common.resources.kaq_redis_resources import KaqQuantRedisRepository
from kaq_quant_common.utils import logger_utils

mutex = threading.Lock()


# redis表监听器
class RedisTableMonitor(MonitorBase):

    def __init__(self, table_name: str, interval=1):
        # 表名
        self._table_name = table_name
        # 执行间隔
        self._interval = interval
        # logger
        self._logger = logger_utils.get_logger(self)
        #
        super().__init__()

    # 这种监听器是辅助型的，而且会开启一个线程，主线程不需要阻塞
    def start(self):
        return super().start(False)

    # ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ abstract methods
    def _do_init(self):
        # 初始化redis
        self._redis = self._init_redis()
        # 查询一次
        self._value = self._do_query()

    def _do_start(self):
        # 开启监听
        self._start_monitor()

    def _do_stop(self):
        # 停止
        self._stop_monitor()

    # ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓
    # 初始化redis
    def _init_redis(self) -> KaqQuantRedisRepository:
        """
        创建redis连接
        """
        redis_config = self._on_get_redis_config()
        host, port, passwd = redis_config["host"], redis_config["port"], redis_config["passwd"]
        return KaqQuantRedisRepository(host=host, port=port, password=passwd)

    # 开启订阅
    def _start_monitor(self):
        # 开启一条线程，定时执行redis查询
        self._ticker_thread_event = threading.Event()

        def query_redis():
            # 上次查询时间
            last_query_time = 0
            while True:
                # 检查是否需要退出
                if self._ticker_thread_event.is_set():
                    self._logger.info("ticker thread exit")
                    break

                # 当前时间
                current_time = time.time()
                # 如果上次请求时间距离当前时间不足，等待
                if current_time - last_query_time < self._interval:
                    time.sleep(0.1)
                    continue

                #
                last_query_time = time.time()

                # self._logger.debug('tick start')
                try:
                    value = self._do_query()

                    # 对比前后的数值
                    if not self._do_compare(value, self._value):
                        # 有变化
                        self._value = value
                        self._on_value_change(value)
                except Exception as e:
                    self._logger.error(f"redis query error: {e}")
                # self._logger.debug('tick finish')
                #
                time.sleep(0.1)

        # 开启线程
        self._ticker_thread = threading.Thread(target=query_redis)
        # 设置为守护线程
        self._ticker_thread.daemon = True
        self._ticker_thread.start()

    def _stop_monitor(self):
        if self._ticker_thread_event is not None:
            self._ticker_thread_event.set()

    # ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ 子类可以重写
    @abstractmethod
    def _do_query(self) -> str:
        # 默认是获取字符串
        str = self._redis.get(self._table_name)
        return str

    @abstractmethod
    def _do_compare(self, value1: any, value2: any) -> bool:
        # 默认认为是字符串
        return value1 == value2

    # ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ abstract methods
    # 需要返回redis配置，包含host, port, passwd 添加类型提示
    @abstractmethod
    def _on_get_redis_config(self) -> dict:
        pass

    @abstractmethod
    def _on_value_change(self, value):
        pass



import threading
import traceback
from string import Template

import dolphindb as ddb
from kaq_quant_common.utils import yml_utils
from kaq_quant_common.utils.logger_utils import get_logger

mutex = threading.Lock()


class KaqQuantDdbStreamInitRepository:
    '''
    定义 asof_join的级联方式, 合并数据到一起, 然后可以订阅判断
    '''

    def __init__(self, host, port, user, passwd, ddb_script_files={}):
        self.logger = get_logger(self)
        if not host or not port or not user or not passwd:
            self.logger.error("【DDB连接参数错误】请检查配置文件")
        if not ddb_script_files or len(ddb_script_files) == 0:
            self.logger.error("【DDB脚本文件列表为空】请检查配置文件")

        mutex.acquire()
        '''
        创建ddb连接 && 添加ddb流数据表支持
        '''
        self.session = ddb.session(enableASYNC=True)
        self.session.connect(host, port, user, passwd)
        self.session.enableStreaming()

        '''
        创建流数据表 && 创建引擎
        '''
        for script_file, args in ddb_script_files.items():
            try:
                # 读取脚本 并转换为gbk编码
                with open(file=script_file, mode="r", encoding="utf-8") as fp:
                    script = fp.read().encode('gbk').decode('gbk')
                if args is None or len(args) == 0:
                    # self.session.runFile(script_file)
                    pass
                else:
                    template = Template(script)
                    script = template.substitute(**args)
                self.session.run(script)
            except Exception as e:
                self.logger.error(f"【创建ddb数据流引擎】 {script_file} 错误异常: {str(e)} - {str(traceback.format_exc())}")
                self.cancel_subscribe()
        mutex.release()

    def cancel_subscribe(self, file_path=None):
        # return
        '''
        1、取消订阅
        PS: 此处不能使用python的localhost
        创建流订阅与引擎
        使用的是ddb的脚步
        传输到服务器上运行
        会调用的节点为服务器ip、port、节点名称
        '''
        if file_path is None or file_path == "":
            self.logger.info("【取消订阅】请传参数file_path")
            return
        self.session.runFile(file_path)

    def drop_streaming(self, file_path=None):
        # return
        '''
        1、取消订阅
        PS: 此处不能使用python的localhost
        创建流订阅与引擎
        使用的是ddb的脚步
        传输到服务器上运行
        会调用的节点为服务器ip、port、节点名称
        '''
        if file_path is None or file_path == "":
            self.logger.info("【取消订阅】请传参数file_path")
            return
        self.session.runFile(file_path)


if __name__ == '__main__':
    host, port, user, passwd = yml_utils.get_ddb_info(pkg_name='kaq_binance_quant')
    ddb_script_files = ['binance_volume_ddb_script.dos', 'binance_limit_order_ddb_script.dos', 'binance_premium_ddb_script.dos']
    ddb_script_files = [yml_utils.get_pkg_file(None, x) for x in ddb_script_files]
    kaq = KaqQuantDdbStreamInitRepository(host, port, user, passwd, ddb_script_files)
    # kaq.drop_streaming()



import threading
import traceback
from abc import abstractmethod

import dolphindb as ddb
import numpy as np
from kaq_quant_common.common.monitor_base import MonitorBase
from kaq_quant_common.utils import logger_utils

mutex = threading.Lock()


# ddb表订阅监听器
class DdbTableMonitor(MonitorBase):

    def __init__(self, table_name: str, action_name: str, batch_size=1000, filter=[]):
        # 表名
        self._table_name = table_name
        #
        self._action_name = action_name
        #
        self._batch_size = batch_size
        #
        self._filter = filter

        # logger
        self._logger = logger_utils.get_logger(self)
        #
        super().__init__()

    # ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ abstract methods
    def _do_init(self):
        # 初始化ddb
        self._init_ddb()

    def _do_start(self):
        # 开启ddb订阅
        self._start_subscribe()

    def _do_stop(self):
        # 关闭订阅
        self._stop_subscribe()

    # ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓
    # 初始化ddb
    def _init_ddb(self):
        '''
        创建ddb连接 && 添加ddb流数据表支持
        '''
        try:
            ddb_config = self._on_get_ddb_config()
            host, port, user, passwd = ddb_config['host'], ddb_config['port'], ddb_config['user'], ddb_config['passwd']
            mutex.acquire()
            self._session = ddb.session(enableASYNC=True)
            self._host = host
            self._port = port
            self._user = user
            self._passwd = passwd
            self._session.connect(host, port, user, passwd)
            self._session.enableStreaming()
        except Exception as e:
            self._logger.error(f'DdbTableMonitor._init_ddb error: {str(e)} - {str(traceback.format_exc())}')
        finally:
            mutex.release()

    # 开启订阅
    def _start_subscribe(self):
        '''
        订阅ddb表
        '''
        self._session.subscribe(
            self._host,
            self._port,
            self._handle,
            tableName=self._table_name,
            actionName=self._action_name,
            filter=np.array(self._filter),
            offset=-1,
            batchSize=self._batch_size,
            throttle=5,
            msgAsTable=True,
        )
        self._logger.info(f'开始订阅 {self._host}:{self._port} {self._table_name} - {self._action_name}')

    def _stop_subscribe(self):
        # TODO
        # script = f"""
        # existsSubscriptionTopic(,`{self._table_name},`{self._action_name})
        # """
        # exitsTopic = self._session.run(script)
        exitsTopic = True
        if exitsTopic is True:
            self._session.unsubscribe(self._host, self._port, self._table_name, self._action_name)
            self._logger.info(f'取消订阅 {self._table_name} - {self._action_name}')
        if not self._session.isClosed():
            self._session.close()

    # ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ abstract methods
    # 需要返回ddb配置，包含host, port, user, passwd 添加类型提示
    @abstractmethod
    def _on_get_ddb_config(self, data) -> dict:
        pass

    @abstractmethod
    def _handle(self, data):
        pass



from pydantic import BaseModel

from . import InstructionRequestBase, InstructionResponseBase


# 资产信息
class AssetsInfo(BaseModel):
    # 币种
    coin: str
    # 余额
    balance: float


# ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ 查询合约账户余额
class ContractBalanceRequest(InstructionRequestBase):
    # 币种
    coin: str


class ContractBalanceResponse(InstructionResponseBase):
    #
    assets: list[AssetsInfo]



import logging
import os
import sys
from typing import Union

from dagster import get_dagster_logger
from loguru import logger

# 清除所有已添加的处理器（包括默认控制台输出）
logger.remove()
# 自定义日志输出
logger.add(
    # 输出到标准输出（控制台）
    sink=sys.stdout,
    # TODO 配置
    level="INFO",
    #
    # format="{time} {level} [{extra[module]}] {message}",  # 显示绑定的module
    format=(
        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
        "<level>{level}</level> | "
        "<cyan>[{extra[module]}]</cyan>"
        "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - "
        "<level>{message}</level>"
    ),
    # 启用颜色（默认已启用，显式指定更清晰）
    colorize=True,
)


def _is_dagster_env():
    # 方法1: 检查Dagster特有的环境变量（推荐）
    # Dagster运行时会设置多个以DAGSTER_开头的环境变量
    is_dagster_env = any(key.startswith("DAGSTER_") for key in os.environ)
    return is_dagster_env


def get_logger(obj: Union[str, object] = None):
    '''
    获取logger
    '''

    is_dagster_env = _is_dagster_env()

    name = obj or ""
    if isinstance(obj, str):
        # do nothing
        name = obj
    elif hasattr(obj, '__class__'):
        name = obj.__class__.__name__
    elif hasattr(obj, '__name__'):
        name = obj.__name__
    else:
        name = ""

    l = None
    if is_dagster_env:
        # TODO
        l = get_dagster_logger()
    else:
        # 使用loguru, 绑定module
        l = logger.bind(module=name)
    return l


########################
# 创建一个适配器，将logging日志转发到Loguru
class LoguruHandler(logging.Handler):

    def emit(self, record):
        # 获取Loguru对应的日志级别
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # 获取日志消息
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        # 转发到Loguru
        logger.opt(depth=depth, exception=record.exc_info).bind(module=record.module).log(level, record.getMessage())


# loguru接管logging
def takeover_logging():
    is_dagster_env = _is_dagster_env()
    if is_dagster_env:
        # 不用处理
        return
    #
    logging.basicConfig(handlers=[LoguruHandler()], level=logging.DEBUG)


takeover_logging()



   
from dagster import DagsterRunStatus, RunsFilter


def dagster_job_check(context, job_name, jobs_number=1):
     # 检查是否有正在运行的相同job
    instance = context.instance
    runs = instance.get_runs(filters=RunsFilter(job_name=job_name, statuses=[DagsterRunStatus.STARTED, DagsterRunStatus.QUEUED]))
    # 过滤出当前job且状态是STARTED的运行
    active_runs = [run for run in runs if run.job_name == job_name and run.status == DagsterRunStatus.STARTED]
    if len(active_runs) > jobs_number:
        # 有正在运行的实例，跳过此次执行
        context.log.info(f'【正在运行:{job_name}.当前任务列表】{str(active_runs)}')
        return False
    context.log.info(f'【当前任务名称】{str(job_name)}')
    return True

def dagster_sensor_check(context, job_name, jobs_number=1):
     # 检查是否有正在运行的相同job
    instance = context.instance
    runs = instance.get_runs(filters=RunsFilter(job_name=job_name, statuses=[DagsterRunStatus.STARTED, DagsterRunStatus.STARTING, DagsterRunStatus.QUEUED]))
    # 过滤出当前job且状态是STARTED的运行
    active_runs = [run for run in runs if run.job_name == job_name]
    if len(active_runs) > jobs_number:
        # 有正在运行的实例，跳过此次执行
        context.log.info(f'【正在运行:{job_name}.当前任务列表】{str(active_runs)}')
        return False
    context.log.info(f'【当前任务名称】{str(job_name)}')
    return True


from enum import Enum
from typing import Optional

from kaq_quant_common.utils import uuid_utils
from pydantic import BaseModel

from . import InstructionRequestBase, InstructionResponseBase


# 订单类型
class OrderType(str, Enum):
    # 现货
    SPOT = "spot"
    # 合约
    FUTURES = "futures"


# 订单方向
class OrderSide(str, Enum):
    BUY = "buy"
    SELL = "sell"


# 持仓方向
class OrderPositionSide(str, Enum):
    # 看多
    LONG = "long"
    # 看空
    SHORT = "short"


# 订单交易类型
class OrderTradeType(str, Enum):
    # 市价
    MARKET = "market"
    # 限价
    LIMIT = "limit"


# 订单状态
class OrderStatus(str, Enum):
    CREATE = "create"
    FINISH = "finish"


# 持仓状态
class PositionStatus(str, Enum):
    OPEN = "open"
    CLOSE = "close"


# 订单信息
class OrderInfo(BaseModel):
    # 指令id-两个交易所会有一笔一样的指令id
    instruction_id: Optional[str] = None
    # 订单id-指定生成的订单id, 如果没有，需要生成
    order_id: Optional[str] = None
    # 交易对
    symbol: str
    # 订单类型 现货/合约
    order_type: OrderType
    # 卖买方向
    side: OrderSide
    # 持仓方向
    position_side: OrderPositionSide
    # 保证金，？什么时候用
    margin: Optional[float] = 0.0
    # 补充保证金，？什么时候用
    supply_margin: Optional[float] = 0.0
    # 杠杆
    level: int
    # 数量(币种，不是USDT)
    quantity: float
    # 限价单才用
    target_price: float
    # 当前价格，？什么时候用
    current_price: Optional[float] = 0.0
    # 交易类型 市价单/限价单
    trade_type: OrderTradeType
    # 风险等级
    risk_level: int
    # 是否强制平仓
    forced_liqu: bool
    # 有效期
    validity_period: Optional[str] = None
    # 策略类型
    strategy_type: Optional[str] = None

    def __init__(self, **data):
        super().__init__(**data)
        if self.instruction_id is None:
            self.instruction_id = uuid_utils.generate_uuid()
        if self.order_id is None:
            self.order_id = uuid_utils.generate_uuid()


# 修改订单信息
class ModifyOrderInfo(BaseModel):
    # 订单ID，自定义的订单id
    order_id: str
    # TODO 暂时不知道修改什么，先定个修改数量
    quantity: Optional[float] = None
    # TODO 暂时不知道修改什么，先定个修改价格
    price: Optional[float] = None


# 已下单信息
class OpenedOrderInfo(BaseModel):
    # 订单ID
    order_id: str
    # 交易对
    symbol: Optional[str] = None
    # 价格
    price: Optional[float] = 0


# ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ 批量下单
# 下单请求
class OrderRequest(InstructionRequestBase):
    orders: list[OrderInfo]


# 下单响应
class OrderResponse(InstructionResponseBase):
    # 返回成功下单的订单信息
    orders: list[OpenedOrderInfo]


# ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ 批量修改订单


class ModifyOrderRequest(InstructionRequestBase):
    orders: list[ModifyOrderInfo]


class ModifyOrderResponse(InstructionResponseBase):
    # 返回成功下单的订单信息
    orders: list[OpenedOrderInfo]


# ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ 撤销订单


class CancelOrderRequest(InstructionRequestBase):
    # 可选的，可能某些平台需要指定交易对
    symbol: Optional[str]
    # 订单id列表
    orders: list[str]


class CancelOrderResponse(InstructionResponseBase):
    orders: list[OpenedOrderInfo]


# ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ 查询当前全部挂单请求
class AllOpenOrdersRequest(InstructionRequestBase):
    # 交易对，用作筛选，不传取全部
    symbol: Optional[str] = None


# 查询当前全部挂单响应
class AllOpenOrdersResponse(InstructionResponseBase):
    # TODO
    orders: list[OpenedOrderInfo]


# ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ 调整杠杆
class ChangeLeverageRequest(InstructionRequestBase):
    # 交易对
    symbol: str
    # 杠杆
    level: int


class ChangeLeverageResponse(InstructionResponseBase):
    # 交易对
    symbol: str
    # 杠杆
    level: int


# ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ 联合保证金模式
# 查询
class QueryMarginModeRequest(InstructionRequestBase):
    pass


class QueryMarginModeResponse(InstructionResponseBase):
    # 是否联合
    is_margin: bool


# 修改
class ChangeMarginModeRequest(InstructionRequestBase):
    # 是否联合
    is_margin: bool


class ChangeMarginModeResponse(InstructionResponseBase):
    # 是否联合
    is_margin: bool



from typing import Optional, Type, TypeVar

import requests
from kaq_quant_common.api.common.auth import get_auth_token
from kaq_quant_common.utils import logger_utils
from pydantic import BaseModel

R = TypeVar("R", bound=BaseModel)


class ApiClientBase:
    """
    api 客户端
    """

    def __init__(self, base_url: str, token: Optional[str] = None):
        self._base_url = base_url.rstrip("/")
        self._token = token if token is not None else get_auth_token()
        self._logger = logger_utils.get_logger(self)

    # 发送请求
    def _make_request(self, method_name: str, request_data: BaseModel, response_model: Type[R]) -> R:
        url = f"{self._base_url}/api/{method_name}"
        headers = {}
        if self._token:
            headers["Authorization"] = f"Bearer {self._token}"
        try:
            # 发送post请求
            response = requests.post(url, json=request_data.model_dump(), headers=headers or None)
            # 检查响应状态码，如果不成功，则尝试解析错误信息并抛出异常
            if not response.ok:
                try:
                    error_data = response.json()
                    error_message = error_data.get("error", response.text)
                except ValueError:
                    error_message = response.text
                raise requests.exceptions.HTTPError(f"HTTP error occurred: {response.status_code} - {error_message}", response=response)
            # 返回请求结果
            return response_model(**response.json())
        except requests.exceptions.RequestException as e:
            self._logger.error(f"An error occurred: {e}")
            raise



import asyncio
import signal
import threading
import time
from typing import Dict, Optional, Set, Tuple

import websockets
from websockets.server import WebSocketServerProtocol

from kaq_quant_common.api.common.api_interface import ApiInterface
from kaq_quant_common.api.common.auth import verify_ws_handshake
from kaq_quant_common.api.ws.models import WsEnvelope, WsMessageType
from kaq_quant_common.utils import logger_utils, signal_utils, uuid_utils


class WsServerBase:
    """
    WebSocket 服务器基类：
    - 动态发现并分派 @api_method 标注的接口方法
    - 维护连接集合与订阅主题
    - 支持请求/响应与服务器端推送
    - 心跳（处理 ping/pong 消息）
    """

    def __init__(self, api: ApiInterface, host: str = "0.0.0.0", port: int = 8765):
        self._api = api
        self._host = host
        self._port = port
        self._logger = logger_utils.get_logger(self)
        # method 映射：name -> (func, request_model, response_model)
        self._api_methods: Dict[str, Tuple] = self._discover_api_methods()
        # 连接管理
        self._connections: Dict[str, WebSocketServerProtocol] = {}
        # 订阅
        self._subscriptions: Dict[str, Set[str]] = {}  # topic -> set(connection_id)

        # 事件循环/服务器控制
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._server: Optional[websockets.server.Server] = None
        self._server_thread: Optional[threading.Thread] = None

    # 动态注册消息处理函数
    def _discover_api_methods(self) -> Dict[str, Tuple]:
        import inspect

        methods = {}
        #
        for name, func in inspect.getmembers(self._api, predicate=inspect.ismethod):
            if hasattr(func, "_is_api_method") and func._is_api_method:
                methods[name] = (func, func._request_model, func._response_model)
        return methods

    # 子类可重写以包装响应（例如添加时间戳）
    def _wrap_response(self, rsp: any):
        return rsp

    # 新的连接
    async def _handle_connection(self, ws: WebSocketServerProtocol):
        connection_id = f"c_{uuid_utils.generate_uuid()}"
        # 创建id
        self._connections[connection_id] = ws
        self._logger.info(f"WS 新连接: {connection_id}")
        try:
            # 接收消息
            async for message in ws:
                try:
                    # 校验消息结构是否正确
                    env = WsEnvelope.model_validate_json(message)
                except Exception as e:
                    self._logger.error(f"解析消息失败: {e}")
                    await ws.send(WsEnvelope(type=WsMessageType.ACK, payload={"error": "invalid message"}).model_dump_json())
                    continue

                # 记录收到的消息类型
                try:
                    self._logger.debug(f"收到消息 type={env.type} method={env.method} topic={env.topic}")
                except Exception:
                    pass

                if env.type == WsMessageType.REQUEST:
                    # 请求->响应
                    await self._on_request(ws, connection_id, env)
                elif env.type == WsMessageType.SUBSCRIBE:
                    # 订阅
                    topic = env.topic or ""
                    if topic:
                        self._subscriptions.setdefault(topic, set()).add(connection_id)
                        await ws.send(WsEnvelope(type=WsMessageType.ACK, payload={"subscribed": topic}).model_dump_json())
                elif env.type == WsMessageType.UNSUBSCRIBE:
                    # 取消订阅
                    topic = env.topic or ""
                    if topic and topic in self._subscriptions and connection_id in self._subscriptions[topic]:
                        self._subscriptions[topic].discard(connection_id)
                        await ws.send(WsEnvelope(type=WsMessageType.ACK, payload={"unsubscribed": topic}).model_dump_json())
                elif env.type == WsMessageType.PING:
                    # 心跳
                    await ws.send(WsEnvelope(type=WsMessageType.PONG).model_dump_json())
                # PUSH 从客户端来的推送默认忽略，按需在子类扩展
        except Exception as e:
            self._logger.warning(f"WS 连接异常 {connection_id}: {e}")
        finally:
            self._connections.pop(connection_id, None)
            # 清理订阅
            for topic, conn_set in self._subscriptions.items():
                conn_set.discard(connection_id)
            self._logger.info(f"WS 连接关闭: {connection_id}")

    # 处理请求，返回响应
    async def _on_request(self, ws: WebSocketServerProtocol, connection_id: str, env: WsEnvelope):
        method_name = env.method or ""
        self._logger.debug(f"处理请求: method={method_name} req_id={env.req_id}")
        req_id = env.req_id

        if method_name not in self._api_methods:
            # 找不到消息处理函数 返回404
            rsp = WsEnvelope(
                type=WsMessageType.RESPONSE,
                req_id=req_id,
                error={"code": 404, "message": f"Method '{method_name}' not found"},
            )
            try:
                await ws.send(rsp.model_dump_json())
            except Exception as e:
                self._logger.error(f"发送404响应失败: {e}")
            return

        # 获取处理函数，请求，返回结构
        method, request_model, response_model = self._api_methods[method_name]
        try:
            # 解析model
            request_obj = request_model(**(env.payload or {}))
            # 执行处理函数，获取返回数据
            response_obj = method(request_obj)
            # 校验响应类型
            if not isinstance(response_obj, response_model):
                raise TypeError("Invalid response type from method")

            # wrap
            response_obj = self._wrap_response(response_obj)
            # 数据
            payload = response_obj.model_dump()
            # 请求返回外层结构，RESPONSE 并且附带 req_id
            rsp = WsEnvelope(type=WsMessageType.RESPONSE, req_id=req_id, payload=payload)
            # 发送消息
            await ws.send(rsp.model_dump_json())
        except Exception as e:
            # 异常处理
            self._logger.error(f"处理请求失败: {e}")
            rsp = WsEnvelope(
                type=WsMessageType.RESPONSE,
                req_id=req_id,
                error={"code": 500, "message": str(e)},
            )
            try:
                await ws.send(rsp.model_dump_json())
            except Exception as send_e:
                self._logger.error(f"发送错误响应失败: {send_e}")

    async def _ws_handler(self, ws: WebSocketServerProtocol):
        # websockets v12+ 新实现的 handler 仅接收 websocket 一个参数
        try:
            # 简单鉴权（若未配置 token 则默认放行）
            ok, err = verify_ws_handshake(getattr(ws, "path", None), getattr(ws, "request_headers", {}))
            if not ok:
                self._logger.warning(f"WS 鉴权失败: {err}")
                try:
                    await ws.close(code=1008, reason=err or "Unauthorized")
                except Exception:
                    pass
                return
            await self._handle_connection(ws)
        except Exception as e:
            # 捕获顶层异常，避免直接导致连接以1011关闭
            self._logger.error(f"WS handler 异常: {e}")

    # 广播消息
    def broadcast(self, topic: str, payload: dict):
        """广播推送到订阅了topic的连接"""
        if not self._loop:
            return
        env = WsEnvelope(type=WsMessageType.PUSH, topic=topic, payload=payload)
        data = env.model_dump_json()

        async def _do_broadcast():
            conns = []
            for conn_id, ws in list(self._connections.items()):
                subs = self._subscriptions.get(topic, set())
                if conn_id in subs:
                    conns.append(ws)
            for ws in conns:
                try:
                    await ws.send(data)
                except Exception as e:
                    self._logger.warning(f"广播失败: {e}")

        asyncio.run_coroutine_threadsafe(_do_broadcast(), self._loop)

    # 定向消息
    def send_to(self, connection_id: str, topic: str, payload: dict):
        """向指定连接推送消息"""
        if not self._loop:
            return
        ws = self._connections.get(connection_id)
        if not ws:
            return
        env = WsEnvelope(type=WsMessageType.PUSH, topic=topic, payload=payload)
        data = env.model_dump_json()

        async def _do_send():
            try:
                await ws.send(data)
            except Exception as e:
                self._logger.warning(f"点对点推送失败: {e}")

        asyncio.run_coroutine_threadsafe(_do_send(), self._loop)

    def run(self):
        """在当前线程启动服务器，阻塞运行"""
        self._logger.info(f"Starting WS server on {self._host}:{self._port}")
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        self._loop = loop

        try:

            async def _start():
                self._server = await websockets.serve(self._ws_handler, self._host, self._port, ping_interval=None)

            # 在事件循环运行中启动服务器（await serve）
            loop.run_until_complete(_start())
            # 运行事件循环直至停止
            loop.run_forever()
        finally:
            # 关闭服务器
            if self._server:
                self._server.close()
                try:
                    loop.run_until_complete(self._server.wait_closed())
                except Exception:
                    pass
            self._logger.warning("WS Server 事件循环结束")
            try:
                loop.close()
            except Exception:
                pass

    def shutdown(self):
        """触发停止事件，优雅关闭服务器"""
        self._logger.info(f"Shutting down WS server on {self._host}:{self._port}")
        if self._loop:
            # 停止事件循环，run() 中的 run_forever 将退出
            try:
                self._loop.call_soon_threadsafe(self._loop.stop)
            except Exception:
                pass

    def run_with_thread(self, block: bool = True):
        self._server_thread = threading.Thread(target=self.run, name="WsServerThread", daemon=True)
        self._server_thread.start()
        # 等待服务器准备就绪的短暂时间
        # 在更复杂场景可使用事件通知ready
        time.sleep(0.5)

        if block:
            self.wait_for_termination()

    def wait_for_termination(self):
        # 终止信号监听
        exit_event = threading.Event()

        def handle_terminate_signal(signum, frame=None):
            self._logger.info(f"收到终止信号 {signum}")
            exit_event.set()
            self.shutdown()

        signal_utils.register_signal_handler(handle_terminate_signal)

        while not exit_event.is_set():
            time.sleep(1)

        self._logger.warning("WsServer 线程退出")

    def shutdown_with_thread(self):
        try:
            self.shutdown()
        finally:
            if hasattr(self, "_server_thread") and self._server_thread and self._server_thread.is_alive():
                self._server_thread.join(timeout=3)



import os
from typing import Any, Optional, Tuple

from kaq_quant_common.utils import yml_utils

# 统一的简单鉴权：基于一个共享的 token
# - 来源优先级：环境变量 KAQ_API_TOKEN > 配置文件 kaq.api_token
# - 如果未配置 token，则认为鉴权关闭（通过）


def _get_expected_token(pkg_name: Optional[str] = None) -> Optional[str]:
    # 环境变量优先
    token = os.getenv("KAQ_API_TOKEN")
    if token:
        return token.strip()
    # 配置文件兜底
    try:
        token_cfg = yml_utils.get(pkg_name or "kaq_quant_common", "api_token")
        if isinstance(token_cfg, str) and token_cfg.strip():
            return token_cfg.strip()
    except Exception as e:
        # 配置读取失败则视为未配置
        pass
    return None


def _extract_token_from_authorization(auth_header: Optional[str]) -> Optional[str]:
    if not auth_header:
        return None
    # 支持 Bearer/Token 两种前缀
    val = auth_header.strip()
    lower = val.lower()
    if lower.startswith("bearer "):
        return val[7:].strip()
    if lower.startswith("token "):
        return val[6:].strip()
    # 若为纯 token 也接受
    return val


def _extract_token_from_headers(headers: Any) -> Optional[str]:
    # 兼容 Flask/Werkzeug 和 websockets.Headers
    try:
        # Flask/requests：大小写不敏感
        auth = headers.get("Authorization") if hasattr(headers, "get") else None
        if not auth and hasattr(headers, "get"):
            auth = headers.get("authorization")
        token = _extract_token_from_authorization(auth)
        if token:
            return token
        # 备用头
        x_token = headers.get("X-API-Token") if hasattr(headers, "get") else None
        if not x_token and hasattr(headers, "get"):
            x_token = headers.get("x-api-token")
        return x_token.strip() if isinstance(x_token, str) and x_token.strip() else None
    except Exception:
        return None


def _extract_token_from_path_query(path: Optional[str]) -> Optional[str]:
    # 解析 ?token=xxx 简单查询参数
    if not path or "?" not in path:
        return None
    try:
        query = path.split("?", 1)[1]
        for pair in query.split("&"):
            if not pair:
                continue
            k, _, v = pair.partition("=")
            if k.lower() == "token" and v:
                return v
    except Exception:
        pass
    return None


# ----------------- 对外校验方法 -----------------


def verify_http_request(flask_request: Any, pkg_name: Optional[str] = None) -> Tuple[bool, Optional[str]]:
    """校验 HTTP 请求头中的 token。
    返回 (是否通过, 错误信息)。未配置 token 时默认放行。
    """
    expected = _get_expected_token(pkg_name)
    if not expected:
        return True, None
    # 从 Header 提取
    token = _extract_token_from_headers(getattr(flask_request, "headers", None))
    if token and token == expected:
        return True, None
    return False, "Unauthorized"


def verify_ws_handshake(path: Optional[str], headers: Any, pkg_name: Optional[str] = None) -> Tuple[bool, Optional[str]]:
    """校验 WebSocket 握手阶段的路径与请求头。
    支持从 Header 与 ?token=xx 解析。未配置 token 时默认放行。
    """
    expected = _get_expected_token(pkg_name)
    if not expected:
        return True, None
    # Header 优先
    token = _extract_token_from_headers(headers)
    if not token:
        # 再从路径查询参数解析
        token = _extract_token_from_path_query(path)
    if token and token == expected:
        return True, None
    return False, "Unauthorized"


# ----------------- 客户端复用：获取默认 token -----------------


def get_auth_token(pkg_name: Optional[str] = None) -> Optional[str]:
    """供客户端复用的获取默认 token 的方法。
    与服务器端校验使用相同的来源规则。
    """
    return _get_expected_token(pkg_name)



import time as new_time
from datetime import datetime, time, timedelta, timezone
from loguru import logger

yymmdd = '%Y-%m-%d'
yymmdd_hhmm = '%Y-%m-%d %H:%M'
yymmdd_hhmmss = '%Y-%m-%d %H:%M:%S'
yymmdd_hhmmss_long = '%Y%m%d%H%M%S'
yymm = '%Y%m'
yy = '%Y'
yyweek = '%Y%W'

morning_start = time(9, 30, 0)
morning_end = time(11, 30, 0)
afternoon_start = time(13, 0, 0)
afternoon_end = time(15, 0, 0)

def get_day_year(_datetime : datetime=None):
    '''
    @param: datetime类型
    获取日期是一年中的第几天
    '''
    try:
        # 获取当前日期
        today = datetime.today()
        if _datetime is not None:
            today = _datetime
        # 获取当前日期是一年中的第几天
        year = today.timetuple().tm_year
        mon = today.timetuple().tm_mon
        day_of_year = today.timetuple().tm_yday
        return year, mon, day_of_year
    except Exception as e:
        logger.error(f'【date_util.get_day_year】异常, - {str(e)}')

def get_week_year(_datetime : datetime=None):
    '''
    @param: datetime类型
    获取日期是一年中的第几周
    '''
    try:
        # 获取当前日期
        today = datetime.today()
        if _datetime is not None:
            today = _datetime
        # 获取当前日期是一年中的第几周
        # today = datetime.date.today()
        # 使用isocalendar()方法获取年、周数和周内日期
        isocalendar = today.isocalendar()
        # year_week 是一个包含三个元素的元组：年、周数、周内日期
        return isocalendar.year, isocalendar.week
    except Exception as e:
        logger.error(f'【date_util.get_week_year】异常, - {str(e)}')

def get_week_year_by_timestamp(timestamp: int=datetime.now().timestamp()):
    if len(str(timestamp)) == 13:
        timestamp = timestamp / 1000
    _date = datetime.fromtimestamp(timestamp)
    return get_week_year(_date)

def get_mon_year(_datetime : datetime=None):
    '''
    @param: datetime类型
    获取日期是一年中的第几月
    '''
    try:
        # 获取当前日期
        today = datetime.today()
        if _datetime is not None:
            today = _datetime
        return today.year, today.month
    except Exception as e:
        logger.error(f'【date_util.get_week_year】异常, - {str(e)}')

def get_mon_year_by_timestamp(timestamp : int=datetime.now().timestamp()):
    '''
    @param: datetime类型
    获取日期是一年中的第几月
    '''
    if len(str(timestamp)) == 13:
        timestamp = timestamp / 1000
    _date = datetime.fromtimestamp(timestamp)
    return get_mon_year(_date)

def get_timezone_utc_timestamp_ms():
    '''
    获取utc时区时间戳数据
    '''
    utc_now = datetime.now(timezone.utc)
    utc_now = datetime(utc_now.year, utc_now.month, utc_now.day, utc_now.hour, utc_now.minute, utc_now.second, utc_now.microsecond)
    return int(utc_now.timestamp()) * 1000

def get_time_zone_diff():
    '''
    获取与utc的时区差值
    '''
    utc_now = datetime.now(timezone.utc)
    now = datetime.now()
    utc_now = datetime(utc_now.year, utc_now.month, utc_now.day, utc_now.hour, utc_now.minute, utc_now.second, utc_now.microsecond)
    now = datetime(now.year, now.month, now.day, now.hour, now.minute, now.second, now.microsecond)
    diff = int(now.timestamp() - utc_now.timestamp())
    return diff

def get_time_zone_diff_ms():
    
    return get_time_zone_diff() * 1000

def day2str(day, format_str='%Y-%m-%d'):
    return datetime.strftime(day, format_str)


def str2day(day_str, format_str='%Y-%m-%d'):
    return datetime.strptime(day_str, format_str)


def str2timestampe(day_str, format_str='%Y-%m-%d %H:%M:%S'):
    dateTime = new_time.strptime(day_str, format_str)
    timestampe = new_time.mktime(dateTime)
    return int(timestampe)


def time2date(time_float):
    time_float = int(time_float)
    return str2day(str(time_float), yymmdd_hhmmss)


def date2time(date):
    return float(day2str(date, yymmdd_hhmmss_long) + '000')


def get_detail_time(_datetime=datetime.today(), hour=0, minute=0, second=0):
    # 获取任意时刻的时间戳
    today_0 = _datetime - timedelta(hours=_datetime.hour, minutes=_datetime.minute, seconds=_datetime.second)
    today_anytime = today_0 + timedelta(hours=hour, minutes=minute, seconds=second)
    # tsp = today_anytime.timestamp()
    # print('{}  的时间戳是  {}'.format(today_anytime, tsp))
    return today_anytime


def get_datestr_start(date):
    timestampe = str2timestampe(date, yymmdd)
    start_date = datetime.fromtimestamp(timestampe)
    ago = get_detail_time(start_date, 0, 0)
    return int(ago.timestamp())


def get_before_now(now, days=0, seconds=0, minutes=0, hours=0, format=yymmdd_hhmmss):
    # 当前时间往前推30天
    ago = now - timedelta(days=days, seconds=seconds, minutes=minutes, hours=hours)
    return ago.strftime(format)


def get_aftertime(n:int=0):  # 精确到分钟
    """
    获取当前时间往后的时间
    :param n: 当前时间后的n分钟
    :return: 返回当前时间后的时间，精确当分钟
    """
    nowtime = datetime.now().strftime('%Y-%m-%d %H:%M:%S')  # 获取当前时间
    # 将当前时间转换为时间数组
    timeArray1= new_time.strptime(nowtime, "%Y-%m-%d %H:%M:%S")
    print(timeArray1)
    # 将时间转换为时间戳:
    timeStamp = int(new_time.mktime(timeArray1))
    print(timeStamp)
    # 在原来时间戳的基础上加n*60
    timeStamp += (n * 60)
    print(timeStamp)
    # 把timestamp处理之后转换为时间数组，格式化为需要的格式
    timeArray2 = new_time.localtime(timeStamp)
    print(timeArray2)
    aftertime = new_time.strftime("%Y-%m-%d %H:%M", timeArray2)
    print(aftertime)
    new_date = (datetime(timeArray2.tm_year, timeArray2.tm_mon, timeArray2.tm_mday, timeArray2.tm_hour, timeArray2.tm_min))
    return new_date, aftertime

def get_current_hour_minutes_time(current_time, hour, minutes):
    '''
    获取之前的时间
    '''
    creation = datetime(current_time.year, current_time.month, current_time.day, hour, minutes)  # 2023年3月15日 12:00
    return creation

def get_before_date(days=0, seconds=0, minutes=0, hours=0, format=yymmdd_hhmmss):
    now = datetime.fromtimestamp(datetime.now(timezone.utc).timestamp())
    # 当前时间往前推30天
    ago = now - timedelta(days=days, seconds=seconds, minutes=minutes, hours=hours)
    return ago.strftime(format)

def get_before_date_start(days :int=0, format=yymmdd_hhmmss):
    '''
    获取当天开始时间,也就是凌晨时间0点
    '''
    machine_now = datetime.now(timezone.utc).timestamp()
    now = datetime.fromtimestamp(machine_now)
    
    ago = now - timedelta(days=days)
    ago = get_detail_time(ago, 0, 0)
    return ago.strftime(format)

if __name__ == '__main__':
    get_week_year_by_timestamp(datetime.now().timestamp())
    get_detail_time()
    get_week_year()



# 定义 服务器api
import time
from abc import ABC, abstractmethod

from kaq_quant_common.api.common.api_interface import ApiInterface, api_method
from kaq_quant_common.api.rest.api_server_base import ApiServerBase
from kaq_quant_common.api.rest.instruction.helper.order_helper import OrderHelper
from kaq_quant_common.api.rest.instruction.models import InstructionResponseBase
from kaq_quant_common.api.rest.instruction.models.account import (
    ContractBalanceRequest,
    ContractBalanceResponse,
)
from kaq_quant_common.api.rest.instruction.models.order import (
    AllOpenOrdersRequest,
    AllOpenOrdersResponse,
    CancelOrderRequest,
    CancelOrderResponse,
    ChangeLeverageRequest,
    ChangeLeverageResponse,
    ModifyOrderRequest,
    ModifyOrderResponse,
    OrderRequest,
    OrderResponse,
)
from kaq_quant_common.api.rest.instruction.models.transfer import (
    TransferRequest,
    TransferResponse,
)
from kaq_quant_common.resources.kaq_mysql_resources import KaqQuantMysqlRepository
from kaq_quant_common.resources.kaq_redis_resources import KaqQuantRedisRepository


class InstructionServerBase(ApiServerBase, ApiInterface, ABC):

    def __init__(self, exchange="", host="0.0.0.0", port=5000):
        super().__init__(self, host, port)
        #
        self._mysql: KaqQuantMysqlRepository = None
        self._redis: KaqQuantRedisRepository = None
        # 交易所
        self._exchange = exchange

        # helper
        self._order_helper = OrderHelper(self)

    # 统一处理返回数据
    def _wrap_response(self, rsp: InstructionResponseBase):
        if rsp is not None:
            # 时间
            if rsp.event_time is None:
                rsp.event_time = int(time.time() * 1000)
        return rsp

    # 下单
    @api_method(OrderRequest, OrderResponse)
    def order(self, request: OrderRequest) -> OrderResponse:
        return self._on_order(request)

    # 修改订单
    @api_method(ModifyOrderRequest, ModifyOrderResponse)
    def modify_order(self, request: ModifyOrderRequest) -> ModifyOrderResponse:
        return self._on_modify_order(request)

    # 取消订单
    @api_method(CancelOrderRequest, CancelOrderResponse)
    def cancel_order(self, request: CancelOrderRequest) -> CancelOrderResponse:
        return self._on_cancel_order(request)

    # 查询当前全部挂单
    @api_method(AllOpenOrdersRequest, AllOpenOrdersResponse)
    def all_open_orders(self, request: AllOpenOrdersRequest) -> AllOpenOrdersResponse:
        return self._on_all_open_orders(request)

    # 调整杠杆
    @api_method(ChangeLeverageRequest, ChangeLeverageResponse)
    def change_leverage(self, request: ChangeLeverageRequest) -> ChangeLeverageResponse:
        return self._on_change_leverage(request)

    # 划转
    @api_method(TransferRequest, TransferResponse)
    def transfer(self, request: TransferRequest) -> TransferResponse:
        return self._on_transfer(request)

    # 查询合约账户余额
    @api_method(ContractBalanceRequest, ContractBalanceResponse)
    def contract_balance(self, request: ContractBalanceRequest) -> ContractBalanceResponse:
        return self._on_contract_balance(request)

    # ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ abstract methods

    @abstractmethod
    def _on_order(self, request: OrderRequest) -> OrderResponse:
        pass

    @abstractmethod
    def _on_modify_order(self, request: ModifyOrderRequest) -> ModifyOrderResponse:
        pass

    @abstractmethod
    def _on_cancel_order(self, request: CancelOrderRequest) -> CancelOrderResponse:
        pass

    @abstractmethod
    def _on_all_open_orders(self, request: AllOpenOrdersRequest) -> AllOpenOrdersResponse:
        pass

    @abstractmethod
    def _on_change_leverage(self, request: ChangeLeverageRequest) -> ChangeLeverageResponse:
        pass

    @abstractmethod
    def _on_transfer(self, request: TransferRequest) -> TransferResponse:
        pass

    @abstractmethod
    def _on_contract_balance(self, request: ContractBalanceRequest) -> ContractBalanceResponse:
        pass



from typing import Callable

from kaq_quant_common.api.ws.ws_client_base import WsClientBase
from kaq_quant_common.api.ws.exchange.models import FundingRateEvent


class WsExchangeClient(WsClientBase):
    """模拟交易所 WS 客户端：资金费率订阅封装"""

    def subscribe_all(self, handler: Callable[[FundingRateEvent], None]):
        def _wrap(payload: dict):
            try:
                evt = FundingRateEvent(**payload)
                handler(evt)
            except Exception:
                # 忽略解析错误，避免阻断接收循环
                pass

        self.subscribe("funding_rate.all", _wrap)

    def subscribe_funding_rate(self, symbol: str, handler: Callable[[FundingRateEvent], None]):
        topic = f"funding_rate.{symbol}"

        def _wrap(payload: dict):
            try:
                evt = FundingRateEvent(**payload)
                handler(evt)
            except Exception:
                pass

        self.subscribe(topic, _wrap)


import time
from abc import ABC, abstractmethod

from kaq_quant_common.api.common.api_interface import ApiInterface, api_method
from kaq_quant_common.api.rest.instruction.models import InstructionResponseBase
from kaq_quant_common.api.rest.instruction.models.account import (
    ContractBalanceRequest,
    ContractBalanceResponse,
)
from kaq_quant_common.api.rest.instruction.models.order import (
    AllOpenOrdersRequest,
    AllOpenOrdersResponse,
    CancelOrderRequest,
    CancelOrderResponse,
    ChangeLeverageRequest,
    ChangeLeverageResponse,
    ChangeMarginModeRequest,
    ChangeMarginModeResponse,
    ModifyOrderRequest,
    ModifyOrderResponse,
    OrderRequest,
    OrderResponse,
    QueryMarginModeRequest,
    QueryMarginModeResponse,
)
from kaq_quant_common.api.rest.instruction.models.position import (
    QueryPositionRequest,
    QueryPositionResponse,
)
from kaq_quant_common.api.rest.instruction.models.transfer import (
    TransferRequest,
    TransferResponse,
)
from kaq_quant_common.api.ws.ws_server_base import WsServerBase


class WsInstructionServerBase(WsServerBase, ApiInterface, ABC):

    def __init__(self, host: str = "0.0.0.0", port: int = 8765):
        super().__init__(self, host, port)

    # 统一处理返回数据
    def _wrap_response(self, rsp: InstructionResponseBase):
        if rsp is not None:
            if rsp.event_time is None:
                rsp.event_time = int(time.time() * 1000)
        return rsp

    # 下单
    @api_method(OrderRequest, OrderResponse)
    def order(self, request: OrderRequest) -> OrderResponse:
        return self._on_order(request)

    # 修改订单
    @api_method(ModifyOrderRequest, ModifyOrderResponse)
    def modify_order(self, request: ModifyOrderRequest) -> ModifyOrderResponse:
        return self._on_modify_order(request)

    # 撤销订单
    @api_method(CancelOrderRequest, CancelOrderResponse)
    def cancel_order(self, request: CancelOrderRequest) -> CancelOrderResponse:
        return self._on_cancel_order(request)

    # 查询当前全部挂单
    @api_method(AllOpenOrdersRequest, AllOpenOrdersResponse)
    def all_open_orders(self, request: AllOpenOrdersRequest) -> AllOpenOrdersResponse:
        return self._on_all_open_orders(request)

    # 调整杠杆
    @api_method(ChangeLeverageRequest, ChangeLeverageResponse)
    def change_leverage(self, request: ChangeLeverageRequest) -> ChangeLeverageResponse:
        return self._on_change_leverage(request)

    # 查询联合保证金模式
    @api_method(QueryMarginModeRequest, QueryMarginModeResponse)
    def query_margin_mode(self, request: QueryMarginModeRequest) -> QueryMarginModeResponse:
        return self._on_query_margin_mode(request)

    # 修改联合保证金模式
    @api_method(ChangeMarginModeRequest, ChangeMarginModeResponse)
    def change_margin_mode(self, request: ChangeMarginModeRequest) -> ChangeMarginModeResponse:
        return self._on_change_margin_mode(request)

    # 查询持仓
    @api_method(QueryPositionRequest, QueryPositionResponse)
    def query_position(self, request: QueryPositionRequest) -> QueryPositionResponse:
        return self._on_query_position(request)

    # 划转
    @api_method(TransferRequest, TransferResponse)
    def transfer(self, request: TransferRequest) -> TransferResponse:
        return self._on_transfer(request)

    # 查询合约账户余额
    @api_method(ContractBalanceRequest, ContractBalanceResponse)
    def contract_balance(self, request: ContractBalanceRequest) -> ContractBalanceResponse:
        return self._on_contract_balance(request)

    # ↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓ abstract methods

    @abstractmethod
    def _on_order(self, request: OrderRequest) -> OrderResponse:
        pass

    @abstractmethod
    def _on_modify_order(self, request: ModifyOrderRequest) -> ModifyOrderResponse:
        pass

    @abstractmethod
    def _on_cancel_order(self, request: CancelOrderRequest) -> CancelOrderResponse:
        pass

    @abstractmethod
    def _on_all_open_orders(self, request: AllOpenOrdersRequest) -> AllOpenOrdersResponse:
        pass

    @abstractmethod
    def _on_change_leverage(self, request: ChangeLeverageRequest) -> ChangeLeverageResponse:
        pass

    @abstractmethod
    def _on_query_margin_mode(self, request: QueryMarginModeRequest) -> QueryMarginModeResponse:
        pass

    @abstractmethod
    def _on_change_margin_mode(self, request: ChangeMarginModeRequest) -> ChangeMarginModeResponse:
        pass

    @abstractmethod
    def _on_query_position(self, request: QueryPositionRequest) -> QueryPositionResponse:
        pass

    @abstractmethod
    def _on_transfer(self, request: TransferRequest) -> TransferResponse:
        pass

    @abstractmethod
    def _on_contract_balance(self, request: ContractBalanceRequest) -> ContractBalanceResponse:
        pass



import signal
import sys

# 1. 为目标信号（例如 SIGINT，即 Ctrl+C 触发）维护一个处理函数列表
signal_handlers = []


# 2. 定义「总处理函数」：触发时依次执行列表中的所有处理逻辑
def total_handler(signalnum, frame):
    # 依次调用所有注册的处理函数
    for handler in signal_handlers:
        handler(signalnum, frame)


# 3. 注册「总处理函数」到目标信号（例如 SIGINT）
# SIGTERM：Dagster通常发送此信号进行终止
# SIGINT：对应Ctrl+C，用于本地测试
signal.signal(signal.SIGTERM, total_handler)
signal.signal(signal.SIGINT, total_handler)


def register_signal_handler(handler):
    signal_handlers.append(handler)



from typing import Optional

from pydantic import BaseModel


class FundingRateEvent(BaseModel):
    # 使用毫秒时间戳，保持与现有事件风格一致
    event_time: int
    exchange_symbol: Optional[str] = None
    exchange: Optional[str] = None
    symbol: Optional[str] = None
    open_rate: Optional[float] = None
    close_rate: Optional[float] = None
    high_rate: Optional[float] = None
    low_rate: Optional[float] = None
    # 毫秒时间戳
    close_time: Optional[int] = None
    # 毫秒时间戳
    next_event_time: Optional[int] = None

    # 兼容表中其他字段
    id: Optional[str] = None
    ctimestamp: Optional[str] = None



import os
import threading
import time
import traceback
from datetime import datetime

import numpy as np
import pandas as pd
from natsort import natsorted
from sqlalchemy import create_engine, text
from sqlalchemy.orm import sessionmaker

from kaq_quant_common.utils import hash_utils, yml_utils
from kaq_quant_common.utils.logger_utils import get_logger

mutex = threading.Lock()


def append_id(df: pd.DataFrame):
    """
    统一添加id和时间戳
    :param df:
    :return:
    """
    if df is not None and not df.empty:
        df = df.copy()
        df["id"] = df.apply(lambda i: hash_utils.generate_hash_id(i), axis=1)

        now_time = time.localtime()
        df["ctimestamp"] = pd.Timestamp(now_time.tm_year, now_time.tm_mon, now_time.tm_mday, now_time.tm_hour, now_time.tm_min, now_time.tm_sec)
    return df


class KaqQuantMysqlRepository:
    """
    获取过往一段时间的归集成交
    数据库创建语句：
    mysql -uroot -pxxxxxxxxxx -e "create database if not exists db_kaq_binance character set 'utf8mb4';"
    PS: 高并发下可能产生错误 (1205, 'Lock wait timeout exceeded; try restarting transaction. 使用`SET GLOBAL innodb_lock_wait_timeout = 3000;`调整)
    同时设置为可以读取已经提交的数据：
    SET SESSION TRANSACTION ISOLATION LEVEL READ COMMITTED;
    SHOW VARIABLES LIKE 'autocommit';
    -- 禁用自动提交, 便于直接一次性提交
    -- SET autocommit = 0;
    """

    def __init__(self, host, port, user, passwd, database, charset="utf8mb4"):
        self.logger = get_logger(self)
        # 创建写入引擎数据库连接池
        self.conn_engine = create_engine(
            f"mysql+mysqldb://{user}:{passwd}@{host}:{port}/{database}?charset={charset}",
            # 连接池大小
            pool_size=30,
            # 超出连接池后，允许的最大扩展数
            max_overflow=5,
            # 池中没有线程最多等待的时间（秒）
            pool_timeout=60,
            # 多久之后，连接自动断开，-1 表示不自动断开（秒）
            pool_recycle=1000,
        )
        # 会话工厂
        self.session_maker = sessionmaker(bind=self.conn_engine, autoflush=False)
        # todo 删掉
        self.session = self.session_maker()

    # 判断表是否存在
    def table_exists(self, table_name):
        # 从连接池获取连接
        session = self.session_maker()
        try:
            tables = session.execute(text("SHOW TABLES")).mappings().all()
            table_list = np.array([[v for _, v in table.items()] for table in tables]).flatten()
            return table_name in table_list
        except Exception as e:
            self.logger.error(f"【mysql-table_exists】异常, {table_name} - {str(e)}")
        finally:
            session.close()

    # 执行sql
    def execute_sql(self, sql, need_commit=False):
        # 从连接池获取连接
        session = self.session_maker()
        ret = None
        try:
            ret = session.execute(text(sql))
            if need_commit:
                session.commit()
        except Exception as e:
            self.logger.error(f"【mysql.execute_sql】异常, {sql} - {str(e)}")
            ret = None
            session.rollback()
        finally:
            session.close()
        return ret

    # 判断表是否存在
    def rename_table(self, table_name):
        # 从连接池获取连接
        session = self.session_maker()
        try:
            time_str = time.strftime("%Y_%m_%d_%H_%M", time.localtime(time.time()))
            sql = f"RENAME TABLE {table_name} TO {table_name}_{time_str};"
            session.execute(text(sql))
            return True
        except Exception as e:
            self.logger.error(f"【mysql-table_exists】异常, {table_name} - {str(e)}")
            return False
        finally:
            session.close()

    def get_table_size(self, table_name, database="db_kaq_binance"):
        """
        查询操作，输入查询语句
        """
        # 从连接池获取连接
        session = self.session_maker()
        try:
            # 从session中获取数据
            sql = f"""
                SELECT ROUND(SUM(data_length + index_length) / 1024 / 1024, 2)
                FROM information_schema.tables 
                WHERE table_name = '{table_name}' AND table_schema = '{database}'
            """
            count = session.execute(text(sql)).fetchone()[0]
            return count
        except Exception as e:
            self.logger.error(f"【mysql-fetch_data】异常, {sql} - {str(e)}")
        finally:
            session.close()
        return 0

    def get_table_list(self):
        # 从连接池获取连接
        session = self.session_maker()
        try:
            # 获取表名
            tables = (
                session.execute(
                    text(
                        "SELECT table_name, create_time FROM information_schema.tables WHERE table_schema = 'db_kaq_binance' ORDER BY create_time DESC"
                    )
                )
                .mappings()
                .all()
            )
            table_list = [table["TABLE_NAME"] for table in tables]
            table_list = list(reversed(natsorted(table_list)))
            return table_list
        except Exception as e:
            self.logger.error(f"【mysql-get_table_list】异常, - {str(e)}")
        finally:
            session.close()
        return []

    def fetch_data(self, query):
        """
        查询操作，输入查询语句
        """
        # 从连接池获取连接
        session = self.session_maker()
        try:
            # 从session中获取数据
            rows = session.execute(text(query)).mappings().all()
            df = pd.DataFrame(rows)
            return df
        except Exception as e:
            self.logger.error(f"【mysql-fetch_data】异常, {query} - {str(e)}")
        finally:
            session.close()
        return pd.DataFrame()

    def fetch_data_count(self, query):
        """
        查询操作，输入查询语句
        """
        # 从连接池获取连接
        session = self.session_maker()
        try:
            # 从session中获取数据
            count = session.execute(text(query)).fetchone()[0]
            return count
        except Exception as e:
            self.logger.error(f"【mysql-fetch_data】异常, {query} - {str(e)}")
        finally:
            session.close()
        return 0

    def get_conn_engine(self):
        """
        create_engine: pandas写入支持的比较好
        """
        return self.conn_engine

    def get_exits_id_list(self, df, table_name):
        """
        获取已经存在的id的列表
        """
        if df is None or df.empty:
            return []
        id_list_str = ", ".join(["'" + _id + "'" for _id in df["id"].values.tolist()])
        id_df = self.fetch_data(f"select id from {table_name} where id in ({id_list_str})")
        exits_id_list = id_df["id"].values.tolist()
        return exits_id_list

    def insert_data(self, df, table_name):
        # 将pdFrame转为table, 并存入mysql
        session = None
        try:
            mutex.acquire(True)
            if df is None or df.empty:
                return
            if "id" not in df:
                df = append_id(df)
            # df = df[~df['id'].isin(self.get_exits_id_list(df, table_name))]
            # pandas写法
            # pd.io.sql.to_sql(df, table_name, self.conn_engine, if_exists='append', index=False, chunksize=100000)

            # 使用 INSERT IGNORE, 如果遇见重复的主键id，则跳过
            columns = ", ".join([_col for _col in df.columns.values])

            # mysql默认一次性可以写入16mb数据，如果超过16mb，有可能阻塞，所以设置为1000条，应该没有问题
            mysql_max_allowed_packet = 300
            total_rows = df.shape[0]
            # 要切分的份数
            part_rows = total_rows // mysql_max_allowed_packet
            if total_rows % mysql_max_allowed_packet > 0:
                part_rows = part_rows + 1
            for i in range(part_rows):
                df_part = df.iloc[i * mysql_max_allowed_packet : (i + 1) * mysql_max_allowed_packet]

                # 拼接插入的sql语句
                value_list = ["(" + ", ".join(["'" + str(_r) + "'" for _r in row.values]) + ")" for index, row in df_part.iterrows()]
                value_list_str = ", ".join(value_list)
                query = f"INSERT IGNORE INTO {table_name} ({columns}) VALUES {value_list_str} ;"
                # 从连接池获取session
                if session is None:
                    session = self.session_maker()
                session.execute(text(query))
            # 一并提交
            if session is not None:
                session.commit()
        except Exception as e:
            self.logger.error(f"【mysql-insert_data异常】- {str(e)} - {str(traceback.format_exc())}")
            if session is not None:
                session.rollback()
        finally:
            mutex.release()
            if session is not None:
                session.close()

    def insert_data_duplicate(self, df, table_name, update_columns: list = None):
        """
        插入数据，如果遇见主键冲突，则更新指定的列
        """
        session = None
        try:
            mutex.acquire(True)
            if df is None or df.empty:
                return
            if "id" not in df:
                df = append_id(df)
            # pandas写法
            # pd.io.sql.to_sql(df, table_name, self.conn_engine, if_exists='append', index=False, chunksize=100000)

            # 使用 INSERT ... ON DUPLICATE KEY UPDATE, 如果遇见重复的主键id，则更新指定的列
            columns = ", ".join([_col for _col in df.columns.values])
            update_columns_str = ", ".join([f"{col}=VALUES({col})" for col in update_columns])

            # mysql默认一次性可以写入16mb数据，如果超过16mb，有可能阻塞，所以设置为1000条，应该没有问题
            mysql_max_allowed_packet = 300
            total_rows = df.shape[0]
            # 要切分的份数
            part_rows = total_rows // mysql_max_allowed_packet
            if total_rows % mysql_max_allowed_packet > 0:
                part_rows = part_rows + 1
            for i in range(part_rows):
                df_part = df.iloc[i * mysql_max_allowed_packet : (i + 1) * mysql_max_allowed_packet]

                # 拼接插入的sql语句
                value_list = ["(" + ", ".join(["'" + str(_r) + "'" for _r in row.values]) + ")" for index, row in df_part.iterrows()]
                value_list_str = ", ".join(value_list)
                query = f"INSERT INTO {table_name} ({columns}) VALUES {value_list_str} ON DUPLICATE KEY UPDATE {update_columns_str} ;"

                # 从连接池获取session
                if session is None:
                    session = self.session_maker()
                session.execute(text(query))
            # 一并提交
            if session is not None:
                session.commit()
        except Exception as e:
            self.logger.error(f"【mysql-insert_data_duplicate异常】- {str(e)} - {str(traceback.format_exc())}")
            if session is not None:
                session.rollback()
        finally:
            mutex.release()
            if session is not None:
                session.close()


def main(query="select * from kaq_binance_perpetual_klines_1h limit 5;"):
    host, port, user, passwd, database, charset = yml_utils.get_mysql_info(os.getcwd())
    KaqQuantMysqlRepository = KaqQuantMysqlRepository(host, port, user, passwd, database, charset)

    print(KaqQuantMysqlRepository.table_exists("kaq_binance_perpetual_klines_1h"))
    # print(KaqQuantMysqlRepository.fetch_data(query).head(3))


# 测试数据库连接池
def test_db_pool():
    host, port, user, passwd, database, charset = yml_utils.get_mysql_info(os.getcwd())
    KaqQuantMysqlRepository = KaqQuantMysqlRepository(host, port, user, passwd, database, charset)

    # 获取一个session
    session1 = KaqQuantMysqlRepository.session_maker()
    session1.execute(text("SELECT 1;")).mappings().all()
    # SHOW PROCESSLIST;  有1条连接

    # 由于上面没有close,这里不会复用上面的连接
    session2 = KaqQuantMysqlRepository.session_maker()
    session2.execute(text("SELECT 1;")).mappings().all()
    # SHOW PROCESSLIST;  有2条连接

    # 关闭session
    session1.close()
    session2.close()

    # 上面session关闭后，这里再创建session会复用之前创建的连接
    session3 = KaqQuantMysqlRepository.session_maker()
    session3.execute(text("SELECT 1;")).mappings().all()
    # SHOW PROCESSLIST;  有2条连接

    print("end")


if __name__ == "__main__":
    main()
